import json
import os
import time

import pytest

from cobbler.utils import get_shared_secret


@pytest.mark.usefixtures("cobbler_xmlrpc_base")
class TestMiscellaneous:
    """
    Class to test remote calls to cobbler which do not belong into a specific category.
    """

    def test_clear_system_logs(self, remote, token, create_kernel_initrd, create_distro, create_profile, create_system,
                               remove_distro, remove_profile, remove_system):
        # Arrange
        fk_kernel = "vmlinuz1"
        fk_initrd = "initrd1.img"
        name_distro = "testdistro_clearsystemlog"
        name_profile = "testprofile_clearsystemlog"
        name_system = "testsystem_clearsystemlog"
        basepath = create_kernel_initrd(fk_kernel, fk_initrd)
        path_kernel = os.path.join(basepath, fk_kernel)
        path_initrd = os.path.join(basepath, fk_initrd)

        distro = create_distro(name_distro, "x86_64", "suse", path_kernel, path_initrd)
        profile = create_profile(name_profile, name_distro, "a=1 b=2 c=3 c=4 c=5 d e")
        system = create_system(name_system, name_profile)

        # Act
        result = remote.clear_system_logs(system, token)

        # Cleanup
        remove_distro(name_distro)
        remove_profile(name_profile)
        remove_system(name_system)

        # Assert
        assert result

    def test_disable_netboot(self, remote, token, create_distro, remove_distro, create_profile, remove_profile,
                             create_system, remove_system, create_kernel_initrd):
        # Arrange
        fk_kernel = "vmlinuz1"
        fk_initrd = "initrd1.img"
        name_distro = "test_distro_template_for_system"
        name_profile = "test_profile_template_for_system"
        name_system = "test_system_template_for_system"
        folder = create_kernel_initrd(fk_kernel, fk_initrd)
        path_kernel = os.path.join(folder, fk_kernel)
        path_initrd = os.path.join(folder, fk_initrd)

        create_distro(name_distro, "x86_64", "suse", path_kernel, path_initrd)
        create_profile(name_profile, name_distro, "text")
        create_system(name_system, name_profile)

        # Act
        result = remote.disable_netboot(name_system, token)

        # Cleanup
        remove_system(name_system)
        remove_profile(name_profile)
        remove_distro(name_distro)

        # Assert
        assert result

    def test_extended_version(self, remote):
        # Arrange

        # Act
        result = remote.extended_version()

        # Assert Example Dict: {'builddate': 'Mon Feb 10 15:38:48 2020', 'gitdate': '?', 'gitstamp': '?', 'version':
        #                       '3.1.2', 'version_tuple': [3, 1, 2]}
        assert type(result) == dict
        assert type(result.get("version_tuple")) == list
        assert [3, 3, 0] == result.get("version_tuple")

    def test_find_items_paged(self, remote, token, create_distro, remove_distro, create_kernel_initrd):
        # Arrange
        fk_kernel = "vmlinuz1"
        fk_initrd = "initrd1.img"
        folder = create_kernel_initrd(fk_kernel, fk_initrd)
        path_kernel = os.path.join(folder, fk_kernel)
        path_initrd = os.path.join(folder, fk_initrd)
        name_distro_1 = "distro_items_paged_1"
        name_distro_2 = "distro_items_paged_2"
        create_distro(name_distro_1, "x86_64", "suse", path_kernel, path_initrd)
        create_distro(name_distro_2, "x86_64", "suse", path_kernel, path_initrd)

        # Act
        result = remote.find_items_paged("distro", None, "name", 1, 1)

        # Cleanup
        remove_distro(name_distro_1)
        remove_distro(name_distro_2)

        # Assert
        # Example output
        # {'items': [{'ctime': 1589386486.9040322, 'depth': 0, 'mtime': 1589386486.9040322, 'source_repos': [],
        # 'tree_build_time': 0, 'uid': 'cbf288465c724c439cf2ede6c94de4e8', 'arch': 'x86_64', 'autoinstall_meta': {},
        # 'boot_files': {}, 'boot_loaders': '<<inherit>>', 'breed': 'suse', 'comment': '', 'fetchable_files': {},
        # 'initrd': '/var/log/cobbler/cobbler.log', 'kernel': '/var/log/cobbler/cobbler.log', 'remote_boot_initrd': '~',
        # 'remote_boot_kernel': '~', 'kernel_options': {}, 'kernel_options_post': {}, 'mgmt_classes': [],
        # 'name': 'distro_items_paged_1', 'os_version': 'virtio26', 'owners': ['admin'], 'redhat_management_key': '',
        # 'template_files': {}}], 'pageinfo': {'page': 1, 'prev_page': '~', 'next_page': 2, 'pages': [1, 2],
        # 'num_pages': 2, 'num_items': 2, 'start_item': 0, 'end_item': 1, 'items_per_page': 1,
        # 'items_per_page_list': [10, 20, 50, 100, 200, 500]}}
        assert type(result) == dict
        assert type(result.get("items")) == list
        assert "pageinfo" in result
        assert "pages" in result["pageinfo"]
        assert result["pageinfo"]["pages"] == [1, 2]

    @pytest.mark.skip("This functionality was implemented very quickly. The test for this needs to be fixed at a "
                      "later point!")
    def test_find_system_by_dns_name(self, remote, token, create_distro, remove_distro, create_profile, remove_profile,
                                     create_system, remove_system, create_kernel_initrd):
        # Arrange
        fk_kernel = "vmlinuz1"
        fk_initrd = "initrd1.img"
        basepath = create_kernel_initrd(fk_kernel, fk_initrd)
        path_kernel = os.path.join(basepath, fk_kernel)
        path_initrd = os.path.join(basepath, fk_initrd)
        name_distro = "test_distro_template_for_system"
        name_profile = "test_profile_template_for_system"
        name_system = "test_system_template_for_system"
        dns_name = "test.cobbler-test.local"
        create_distro(name_distro, "x86_64", "suse", path_kernel, path_initrd)
        create_profile(name_profile, name_distro, "text")
        system = create_system(name_system, name_profile)
        remote.modify_system(system, "dns_name", dns_name, token)
        remote.save_system(system, token)

        # Act
        result = remote.find_system_by_dns_name(dns_name)

        # Cleanup
        remove_system(name_system)
        remove_profile(name_profile)
        remove_distro(name_distro)

        # Assert
        assert result

    def test_generate_script(self, remote, create_distro, remove_distro, create_profile, remove_profile,
                             create_system, remove_system, create_kernel_initrd):
        # Arrange
        fk_kernel = "vmlinuz1"
        fk_initrd = "initrd1.img"
        basepath = create_kernel_initrd(fk_kernel, fk_initrd)
        path_kernel = os.path.join(basepath, fk_kernel)
        path_initrd = os.path.join(basepath, fk_initrd)
        name_distro = "test_distro_template_for_system"
        name_profile = "test_profile_template_for_system"
        name_autoinstall_script = "test_generate_script"
        create_distro(name_distro, "x86_64", "suse", path_kernel, path_initrd)
        create_profile(name_profile, name_distro, "text")
        # TODO: Create Autoinstall Script

        # Act
        result = remote.generate_script(name_profile, None, name_autoinstall_script)

        # Cleanup
        remove_profile(name_profile)
        remove_distro(name_distro)

        # Assert
        assert result

    def test_get_item_as_rendered(self, remote, token, create_distro, remove_distro, create_kernel_initrd):
        # Arrange
        fk_kernel = "vmlinuz1"
        fk_initrd = "initrd1.img"
        basepath = create_kernel_initrd(fk_kernel, fk_initrd)
        path_kernel = os.path.join(basepath, fk_kernel)
        path_initrd = os.path.join(basepath, fk_initrd)
        name = "test_item_as_rendered"
        create_distro(name, "x86_64", "suse", path_kernel, path_initrd)

        # Act
        result = remote.get_distro_as_rendered(name, token)

        # Cleanup
        remove_distro(name)

        # Assert
        assert result

    def test_get_s_since(self, remote, create_distro, remove_distro, create_kernel_initrd):
        # Arrange
        fk_kernel = "vmlinuz1"
        fk_initrd = "initrd1.img"
        basepath = create_kernel_initrd(fk_kernel, fk_initrd)
        path_kernel = os.path.join(basepath, fk_kernel)
        path_initrd = os.path.join(basepath, fk_initrd)
        name_distro_before = "test_distro_since_before"
        name_distro_after = "test_distro_since_after"
        create_distro(name_distro_before, "x86_64", "suse", path_kernel, path_initrd)
        mtime = float(time.time())
        create_distro(name_distro_after, "x86_64", "suse", path_kernel, path_initrd)

        # Act
        result = remote.get_distros_since(mtime)

        # Cleanup
        remove_distro(name_distro_before)
        remove_distro(name_distro_after)

        # Assert
        assert isinstance(result, list)
        assert len(result) == 1

    def test_get_authn_module_name(self, remote, token):
        # Arrange

        # Act
        result = remote.get_authn_module_name(token)

        # Assert
        assert result

    def test_get_blended_data(self, remote, create_distro, remove_distro, create_profile, remove_profile,
                              create_system, remove_system, create_kernel_initrd):
        # Arrange
        fk_kernel = "vmlinuz1"
        fk_initrd = "initrd1.img"
        basepath = create_kernel_initrd(fk_kernel, fk_initrd)
        path_kernel = os.path.join(basepath, fk_kernel)
        path_initrd = os.path.join(basepath, fk_initrd)
        name_distro = "test_distro_blended"
        name_profile = "test_profile_blended"
        name_system = "test_system_blended"
        create_distro(name_distro, "x86_64", "suse", path_kernel, path_initrd)
        create_profile(name_profile, name_distro, "text")
        create_system(name_system, name_profile)

        # Act
        result = remote.get_blended_data(name_profile, name_system)

        # Cleanup
        remove_system(name_system)
        remove_profile(name_profile)
        remove_distro(name_distro)

        # Assert
        assert result

    def test_get_config_data(self, remote, token, create_distro, remove_distro, create_profile, remove_profile,
                             create_system, remove_system, create_kernel_initrd):
        # Arrange
        fk_kernel = "vmlinuz1"
        fk_initrd = "initrd1.img"
        basepath = create_kernel_initrd(fk_kernel, fk_initrd)
        path_kernel = os.path.join(basepath, fk_kernel)
        path_initrd = os.path.join(basepath, fk_initrd)
        name_distro = "test_distro_template_for_system"
        name_profile = "test_profile_template_for_system"
        name_system = "test_system_template_for_system"
        system_hostname = "testhostname"
        create_distro(name_distro, "x86_64", "suse", path_kernel, path_initrd)
        create_profile(name_profile, name_distro, "text")
        system = create_system(name_system, name_profile)
        remote.modify_system(system, "hostname", system_hostname, token)
        remote.save_system(system, token)

        # Act
        result = remote.get_config_data(system_hostname)

        # Cleanup
        remove_system(name_system)
        remove_profile(name_profile)
        remove_distro(name_distro)

        # Assert
        assert json.loads(result)

    def test_get_repos_compatible_with_profile(self, remote, token, create_distro, remove_distro, create_profile,
                                               remove_profile, create_repo, remove_repo, create_kernel_initrd):
        # Arrange
        fk_kernel = "vmlinuz1"
        fk_initrd = "initrd1.img"
        basepath = create_kernel_initrd(fk_kernel, fk_initrd)
        path_kernel = os.path.join(basepath, fk_kernel)
        path_initrd = os.path.join(basepath, fk_initrd)
        name_distro = "test_distro_get_repo_for_profile"
        name_profile = "test_profile_get_repo_for_profile"
        name_repo_compatible = "test_repo_compatible_profile_1"
        name_repo_incompatible = "test_repo_compatible_profile_2"
        create_distro(name_distro, "x86_64", "suse", path_kernel, path_initrd)
        create_profile(name_profile, name_distro, "text")
        repo_compatible = create_repo(name_repo_compatible, "http://localhost", False)
        repo_incompatible = create_repo(name_repo_incompatible, "http://localhost", False)
        remote.modify_repo(repo_compatible, "arch", "x86_64", token)
        remote.save_repo(repo_compatible, token)
        remote.modify_repo(repo_incompatible, "arch", "ppc64le", token)
        remote.save_repo(repo_incompatible, token)

        # Act
        result = remote.get_repos_compatible_with_profile(name_profile, token)

        # Cleanup
        remove_profile(name_profile)
        remove_distro(name_distro)
        remove_repo(name_repo_compatible)
        remove_repo(name_repo_incompatible)

        # Assert
        assert result != []

    def test_get_status(self, remote, token):
        # Arrange

        # Act
        result = remote.get_status("normal", token)

        # Assert
        assert result == {}

    @pytest.mark.skip("The function under test appears to have a bug. For now we skip the test.")
    def test_get_template_file_for_profile(self, remote, create_distro, remove_distro, create_profile, remove_profile,
                                           create_autoinstall_template, remove_autoinstall_template,
                                           create_kernel_initrd):
        # Arrange
        fk_kernel = "vmlinuz1"
        fk_initrd = "initrd1.img"
        basepath = create_kernel_initrd(fk_kernel, fk_initrd)
        path_kernel = os.path.join(basepath, fk_kernel)
        path_initrd = os.path.join(basepath, fk_initrd)
        name_distro = "test_distro_template_for_profile"
        name_profile = "test_profile_template_for_profile"
        name_template = "test_template_for_profile"
        content_template = "# Testtemplate"
        create_distro(name_distro, "x86_64", "suse", path_kernel, path_initrd)
        create_profile(name_profile, name_distro, "text")
        create_autoinstall_template(name_template, content_template)

        # Act
        # TODO: Fix test & functionality!
        result = remote.get_template_file_for_profile(name_profile, name_template)

        # Cleanup
        remove_profile(name_profile)
        remove_distro(name_distro)
        remove_autoinstall_template(name_template)

        # Assert
        assert result == content_template

    def test_get_template_file_for_system(self, remote, create_distro, remove_distro, create_profile, remove_profile,
                                          create_system, remove_system, create_autoinstall_template,
                                          remove_autoinstall_template, create_kernel_initrd):
        # Arrange
        fk_kernel = "vmlinuz1"
        fk_initrd = "initrd1.img"
        basepath = create_kernel_initrd(fk_kernel, fk_initrd)
        path_kernel = os.path.join(basepath, fk_kernel)
        path_initrd = os.path.join(basepath, fk_initrd)
        name_distro = "test_distro_template_for_system"
        name_profile = "test_profile_template_for_system"
        name_system = "test_system_template_for_system"
        name_template = "test_template_for_system"
        content_template = "# Testtemplate"
        create_distro(name_distro, "x86_64", "suse", path_kernel, path_initrd)
        create_profile(name_profile, name_distro, "text")
        create_system(name_system, name_profile)
        create_autoinstall_template(name_template, content_template)

        # Act
        result = remote.get_template_file_for_system(name_system, name_template)

        # Cleanup
        remove_system(name_system)
        remove_profile(name_profile)
        remove_distro(name_distro)
        remove_autoinstall_template(name_template)

        # Assert
        assert result

    def test_is_autoinstall_in_use(self, remote, token, create_distro, remove_distro, create_profile, remove_profile,
                                   create_kernel_initrd):
        # Arrange
        fk_kernel = "vmlinuz1"
        fk_initrd = "initrd1.img"
        basepath = create_kernel_initrd(fk_kernel, fk_initrd)
        path_kernel = os.path.join(basepath, fk_kernel)
        path_initrd = os.path.join(basepath, fk_initrd)
        name_distro = "test_distro_is_autoinstall_in_use"
        name_profile = "test_profile_is_autoinstall_in_use"
        create_distro(name_distro, "x86_64", "suse", path_kernel, path_initrd)
        create_profile(name_profile, name_distro, "text")

        # Act
        result = remote.is_autoinstall_in_use(name_profile, token)

        # Cleanup
        remove_profile(name_profile)
        remove_distro(name_distro)

        # Assert
        assert not result

    def test_logout(self, remote):
        # Arrange
        shared_secret = get_shared_secret()
        newtoken = remote.login("", shared_secret)

        # Act
        resultlogout = remote.logout(newtoken)
        resulttokencheck = remote.token_check(newtoken)

        # Assert
        assert resultlogout
        assert not resulttokencheck

    def test_modify_setting(self, remote, token):
        # Arrange

        # Act
        result = remote.modify_setting("auth_token_expiration", 7200, token)

        # Assert
        assert result == 1

    def test_read_autoinstall_template(self, remote, token, create_autoinstall_template, remove_autoinstall_template):
        # Arrange
        name = "test_template_name"
        create_autoinstall_template(name, "# Testtemplate")

        # Act
        result = remote.read_autoinstall_template(name, token)

        # Cleanup
        remove_autoinstall_template(name)

        # Assert
        assert result

    def test_write_autoinstall_template(self, remote, token, remove_autoinstall_template):
        # Arrange
        name = "testtemplate"

        # Act
        result = remote.write_autoinstall_template(name, "# Testtemplate", token)

        # Cleanup
        remove_autoinstall_template(name)

        # Assert
        assert result

    def test_remove_autoinstall_template(self, remote, token, create_autoinstall_template):
        # Arrange
        name = "test_template_remove"
        create_autoinstall_template(name, "# Testtemplate")

        # Act
        result = remote.remove_autoinstall_template(name, token)

        # Assert
        assert result

    def test_read_autoinstall_snippet(self, remote, token, testsnippet, snippet_add, snippet_remove):
        # Arrange
        snippet_name = "testsnippet_read"
        snippet_add(snippet_name, testsnippet)

        # Act
        result = remote.read_autoinstall_snippet(snippet_name, token)

        # Assert
        assert result == testsnippet

        # Cleanup
        snippet_remove(snippet_name)

    def test_write_autoinstall_snippet(self, remote, token, testsnippet, snippet_remove):
        # Arrange
        # See fixture: testsnippet
        name = "testsnippet_write"

        # Act
        result = remote.write_autoinstall_snippet(name, testsnippet, token)

        # Assert
        assert result

        # Cleanup
        snippet_remove(name)

    def test_remove_autoinstall_snippet(self, remote, token, snippet_add, testsnippet):
        # Arrange
        name = "testsnippet_remove"
        snippet_add(name, testsnippet)

        # Act
        result = remote.remove_autoinstall_snippet(name, token)

        # Assert
        assert result

    def test_run_install_triggers(self, remote, token):
        # Arrange
        # TODO: Needs a system as a target

        # Act
        result_pre = remote.run_install_triggers("pre", "system", "systemname", "10.0.0.2", token)
        result_post = remote.run_install_triggers("post", "system", "systemname", "10.0.0.2", token)

        # Assert
        assert result_pre
        assert result_post

    def test_version(self, remote):
        # Arrange

        # Act
        result = remote.version()

        # Assert
        # Will fail if the version is adjusted in the setup.py
        assert result == 3.3

    def test_xapi_object_edit(self, remote, token, remove_distro, create_kernel_initrd):
        # Arrange
        fk_kernel = "vmlinuz1"
        fk_initrd = "initrd1.img"
        basepath = create_kernel_initrd(fk_kernel, fk_initrd)
        path_kernel = os.path.join(basepath, fk_kernel)
        path_initrd = os.path.join(basepath, fk_initrd)
        name = "testdistro_xapi_edit"

        # Act
        result = remote.xapi_object_edit("distro", name, "add",
                                         {"name": name, "arch": "x86_64", "breed": "suse", "kernel": path_kernel,
                                          "initrd": path_initrd}, token)

        # Cleanup
        remove_distro(name)

        # Assert
        assert result

    @pytest.mark.usefixtures("create_testdistro", "create_testmenu", "create_profile", "remove_testdistro",
                             "remove_testmenu", "remove_testprofile")
    def test_render_vars(self, remote, token):
        """
        Test: string replacements for @@xyz@@
        """

        # Arrange --> There is nothing to be arranged
        kernel_options = "tree=http://@@http_server@@/cblr/links/@@distro_name@@"

        # Act
        distro = remote.get_item_handle("distro", "testdistro0", token)
        remote.modify_distro(distro, "kernel_options", kernel_options, token)
        remote.save_distro(distro, token)

        # Assert --> Let the test pass if the call is okay.
        assert True

    @pytest.mark.skip("Functionality is broken!")
    @pytest.mark.usefixtures("create_testdistro", "create_testmenu", "create_testprofile", "create_testsystem",
                             "remove_testdistro", "remove_testmenu", "remove_testprofile", "remove_testsystem")
    def test_upload_log_data(self, remote):
        # Arrange

        # Act
        result = remote.upload_log_data("testsystem0", "testinstall.log", 0, 0, b"asdas")

        # Assert
        assert isinstance(result, bool)
        assert result



from cobbler.api import CobblerAPI
from cobbler.modules.authentication import pam


class TestPam:
    def test_authenticate(self):
        # Arrange
        test_username = "test"
        test_password = "test"
        test_api = CobblerAPI()

        # Act
        result = pam.authenticate(test_api, test_username, test_password)

        # Assert
        assert result



"""
TODO
"""

import enum

VALUE_INHERITED = "<<inherit>>"
VALUE_NONE = "none"


class ResourceAction(enum.Enum):
    """
    This enum represents all actions a resource may execute.
    """
    CREATE = "create"
    REMOVE = "remove"


class NetworkInterfaceType(enum.Enum):
    """
    This enum represents all interface types Cobbler is able to set up on a target host.
    """
    NA = 0
    BOND = 1
    BOND_SLAVE = 2
    BRIDGE = 3
    BRIDGE_SLAVE = 4
    BONDED_BRIDGE_SLAVE = 5
    BMC = 6
    INFINIBAND = 7


class RepoBreeds(enum.Enum):
    """
    This enum describes all repository breeds Cobbler is able to manage.
    """
    NONE = VALUE_NONE
    RSYNC = "rsync"
    RHN = "rhn"
    YUM = "yum"
    APT = "apt"
    WGET = "wget"


class RepoArchs(enum.Enum):
    """
    This enum describes all repository architectures Cobbler is able to serve in case the content of the repository is
    serving the same architecture.
    """
    NONE = VALUE_NONE
    I386 = "i386"
    X86_64 = "x86_64"
    IA64 = "ia64"
    PPC = "ppc"
    PPC64 = "ppc64"
    PPC64LE = "ppc64le"
    PPC64EL = "ppc64el"
    S390 = "s390"
    ARM = "arm"
    AARCH64 = "aarch64"
    NOARCH = "noarch"
    SRC = "src"


class Archs(enum.Enum):
    """
    This enum describes all system architectures which Cobbler is able to provision.
    """
    I386 = "i386"
    X86_64 = "x86_64"
    IA64 = "ia64"
    PPC = "ppc"
    PPC64 = "ppc64"
    PPC64LE = "ppc64le"
    PPC64EL = "ppc64el"
    S390 = "s390"
    S390X = "s390x"
    ARM = "arm"
    AARCH64 = "aarch64"


class VirtType(enum.Enum):
    """
    This enum represents all known types of virtualization Cobbler is able to handle via Koan.
    """
    INHERTIED = VALUE_INHERITED
    QEMU = "qemu"
    KVM = "kvm"
    XENPV = "xenpv"
    XENFV = "xenfv"
    VMWARE = "vmware"
    VMWAREW = "vmwarew"
    OPENVZ = "openvz"
    AUTO = "auto"


class VirtDiskDrivers(enum.Enum):
    """
    This enum represents all virtual disk driver Cobbler can handle.
    """
    INHERTIED = VALUE_INHERITED
    RAW = "raw"
    QCOW2 = "qcow2"
    QED = "qed"
    VDI = "vdi"
    VDMK = "vdmk"


class BaudRates(enum.Enum):
    """
    This enum describes all baud rates which are commonly used.
    """
    B0 = 0
    B110 = 110
    B300 = 300
    B600 = 600
    B1200 = 1200
    B2400 = 2400
    B4800 = 4800
    B9600 = 9600
    B14400 = 14400
    B19200 = 19200
    B38400 = 38400
    B57600 = 57600
    B115200 = 115200
    B128000 = 128000
    B256000 = 256000


class ImageTypes(enum.Enum):
    """
    This enum represents all image types which Cobbler can manage.
    """
    DIRECT = "direct"
    ISO = "iso"
    MEMDISK = "memdisk"
    VIRT_CLONE = "virt-clone"


class MirrorType(enum.Enum):
    """
    This enum represents all mirror types which Cobbler can manage.
    """
    METALINK = "metalink"
    MIRRORLIST = "mirrorlist"
    BASEURL = "baseurl"



#!/usr/bin/python3

import os
import glob
import simplejson
import subprocess

import cobbler.api as capi


def serialize_item(collection, item):
    """
    Save a collection item to file system

    @param collection name
    @param item dictionary
    """

    filename = "/var/lib/cobbler/collections/%s/%s" % (collection, item['name'])

    if capi.CobblerAPI().settings().serializer_pretty_json:
        sort_keys = True
        indent = 4
    else:
        sort_keys = False
        indent = None

    filename += ".json"
    fd = open(filename, "w+")
    data = simplejson.dumps(item, encoding="utf-8", sort_keys=sort_keys, indent=indent)
    fd.write(data)

    fd.close()

def deserialize_raw_old(collection_types):

    results = []

    all_files = glob.glob("/var/lib/cobbler/config/%s/*" % collection_types)

    for f in all_files:
        fd = open(f)
        json_data = fd.read()
        _dict = simplejson.loads(json_data, encoding='utf-8')
        results.append(_dict)
        fd.close()
    return results


def substitute_paths(value):
    if isinstance(value, list):
        new_value = []
        for item in value:
            new_value.append(substitute_paths(item))
        value = new_value
    elif isinstance(value, str):
        value = value.replace('/ks_mirror/','/distro_mirror/')
    return value

def transform_key(key, value):
    if key in transform:
        ret_value = transform[key](value)
    else:
        ret_value = value

    return substitute_paths(ret_value)

# Keys to add to various collections
add = {
  "distros": {
    "boot_loader": "grub",
  },
  "profiles": {
    "next_server": "<<inherit>>",
  },
  "systems": {
    "boot_loader": "<<inherit>>",
    "next_server": "<<inherit>>",
    "power_identity_file": "",
    "power_options": "",
    "serial_baud_rate": "",
    "serial_device": "",
  },
}

# Keys to remove
remove = [
  "ldap_enabled",
  "ldap_type",
  "monit_enabled",
  "redhat_management_server",
  "template_remote_kickstarts",
]

# Keys to rename
rename = {
  "kickstart": "autoinstall",
  "ks_meta": "autoinstall_meta",
}

# Keys to transform - use new key name if renamed
transform = {
  "autoinstall": os.path.basename,
}

# Convert the old collections to new collections
for old_type in ['distros.d','files.d','images.d','mgmtclasses.d','packages.d','profiles.d','repos.d','systems.d']:
    new_type = old_type[:-2]
    # Load old files
    old_collection = deserialize_raw_old(old_type)
    print("Processing %s:" % old_type)

    for old_item in old_collection:
        print("    Processing %s" % old_item['name'])
        new_item = {}
        for key in old_item:
            if key in remove:
                continue
            if key in rename:
                new_item[rename[key]] = transform_key(rename[key], old_item[key])
                continue
            new_item[key] = transform_key(key, old_item[key])

        if new_type in add:
            new_item.update(add[new_type])

        serialize_item(new_type, new_item)

path_rename = [
  ("/var/lib/cobbler/kickstarts", "/var/lib/cobbler/templates"),
  ("/var/www/cobbler/ks_mirror", "/var/www/cobbler/distro_mirror"),
]

# Copy paths
for old_path, new_path in path_rename:
    if os.path.isdir(old_path):
        subprocess.run(["cp", "-al", "%s/*" % old_path, "%s/" % new_path], shell=True)



import os
import re
import binascii
import tempfile
from typing import Optional

import cobbler.utils as utils
import cobbler.templar as templar
import cobbler.tftpgen as tftpgen
import logging

HAS_HIVEX = True
try:
    import pefile
    import hivex
    from hivex.hive_types import REG_DWORD
    from hivex.hive_types import REG_BINARY
    from hivex.hive_types import REG_SZ
    from hivex.hive_types import REG_MULTI_SZ
except Exception:
    HAS_HIVEX = False

answerfile_template_name = "answerfile.template"
post_inst_cmd_template_name = "post_inst_cmd.template"
startnet_template_name = "startnet.template"
wimupdate = "/usr/bin/wimupdate"

logger = logging.getLogger()


def register() -> Optional[str]:
    """
    This pure python trigger acts as if it were a legacy shell-trigger, but is much faster. The return of this method
    indicates the trigger type
    :return: Always ``/var/lib/cobbler/triggers/sync/post/*``
    """
    if not HAS_HIVEX:
        logging.info("python3-hivex not found. If you need Automatic Windows Installation support, please install.")
        return

    return "/var/lib/cobbler/triggers/sync/post/*"


def bcdedit(orig_bcd, new_bcd, wim, sdi, startoptions=None):
    def winpath_length(wp, add):
        wpl = add + 2 * len(wp)
        return wpl.to_bytes((wpl.bit_length() + 7) // 8, 'big')

    def guid2binary(g):
        guid = g[7] + g[8] + g[5] + g[6] + g[3] + g[4] + g[1] + g[2] + g[12] + g[13] + g[10] + g[11] + g[17] + g[18]
        guid += g[15] + g[16] + g[20] + g[21] + g[22] + g[23] + g[25] + g[26] + g[27] + g[28] + g[29] + g[30] + g[31]
        guid += g[32] + g[33] + g[34] + g[35] + g[36]
        return binascii.unhexlify(guid)

    wim = wim.replace('/', '\\')
    sdi = sdi.replace('/', '\\')

    h = hivex.Hivex(orig_bcd, write=True)
    root = h.root()
    objs = h.node_get_child(root, "Objects")

    for n in h.node_children(objs):
        h.node_delete_child(n)

    b = h.node_add_child(objs, "{9dea862c-5cdd-4e70-acc1-f32b344d4795}")
    d = h.node_add_child(b, "Description")
    h.node_set_value(d, {"key": "Type", "t": REG_DWORD, "value": b"\x02\x00\x10\x10"})
    e = h.node_add_child(b, "Elements")
    e1 = h.node_add_child(e, "25000004")
    h.node_set_value(e1, {"key": "Element", "t": REG_BINARY, "value": b"\x1e\x00\x00\x00\x00\x00\x00\x00"})
    e1 = h.node_add_child(e, "12000004")
    h.node_set_value(e1, {"key": "Element", "t": REG_SZ, "value": "Windows Boot Manager\0".encode(encoding="utf-16le")})
    e1 = h.node_add_child(e, "24000001")
    h.node_set_value(e1, {"key": "Element", "t": REG_MULTI_SZ,
                          "value": "{65c31250-afa2-11df-8045-000c29f37d88}\0\0".encode(encoding="utf-16le")})
    e1 = h.node_add_child(e, "16000048")
    h.node_set_value(e1, {"key": "Element", "t": REG_BINARY, "value": b"\x01"})

    b = h.node_add_child(objs, "{65c31250-afa2-11df-8045-000c29f37d88}")
    d = h.node_add_child(b, "Description")
    h.node_set_value(d, {"key": "Type", "t": REG_DWORD, "value": b"\x03\x00\x20\x13"})
    e = h.node_add_child(b, "Elements")
    e1 = h.node_add_child(e, "12000004")
    h.node_set_value(e1, {"key": "Element", "t": REG_SZ, "value": "Windows PE\0".encode(encoding="utf-16le")})
    e1 = h.node_add_child(e, "22000002")
    h.node_set_value(e1, {"key": "Element", "t": REG_SZ, "value": "\\Windows\0".encode(encoding="utf-16le")})
    e1 = h.node_add_child(e, "26000010")
    h.node_set_value(e1, {"key": "Element", "t": REG_BINARY, "value": b"\x01"})
    e1 = h.node_add_child(e, "26000022")
    h.node_set_value(e1, {"key": "Element", "t": REG_BINARY, "value": b"\x01"})
    e1 = h.node_add_child(e, "11000001")
    guid = guid2binary("{ae5534e0-a924-466c-b836-758539a3ee3a}")
    wimval = {"key": "Element",
              "t": REG_BINARY,
              "value": guid + b"\x00\x00\x00\x00\x01\x00\x00\x00" + winpath_length(wim, 126)
                            + b"\x00\x00\x00\x00\x00\x00\x00\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                              b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00"
                            + winpath_length(wim, 86)
                            + b"\x00\x00\x00\x05\x00\x00\x00\x05\x00\x00\x00\x00\x00\x00\x00\x48\x00\x00"
                              b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                              b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                              b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                              b"\x00\x00\x00\x00\x00\x00\x00" + wim.encode(encoding="utf_16_le")
                            + b"\x00\x00"}
    h.node_set_value(e1, wimval)
    e1 = h.node_add_child(e, "21000001")
    h.node_set_value(e1, wimval)

    if startoptions:
        e1 = h.node_add_child(e, "12000030")
        h.node_set_value(e1,
                         {"key": "Element", "t": REG_SZ, "value": startoptions.join("\0").encode(encoding="utf-16le")})

    b = h.node_add_child(objs, "{ae5534e0-a924-466c-b836-758539a3ee3a}")
    d = h.node_add_child(b, "Description")
    h.node_set_value(d, {"key": "Type", "t": REG_DWORD, "value": b"\x00\x00\x00\x30"})
    e = h.node_add_child(b, "Elements")
    e1 = h.node_add_child(e, "12000004")
    h.node_set_value(e1, {"key": "Element", "t": REG_SZ, "value": "Ramdisk Options\0".encode(encoding="utf-16le")})
    e1 = h.node_add_child(e, "32000004")
    h.node_set_value(e1, {"key": "Element", "t": REG_SZ, "value": sdi.encode(encoding="utf-16le") + b"\x00\x00"})
    e1 = h.node_add_child(e, "31000003")
    h.node_set_value(e1, {"key": "Element",
                          "t": REG_BINARY,
                          "value": b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x05\x00\x00\x00"
                                   b"\x00\x00\x00\x00\x48\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                                   b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                                   b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                                   b"\x00\x00\x00\x00\x00\x00\x00\x00"})
    h.commit(new_bcd)


def run(api, args):
    settings = api.settings()
    if not settings.windows_enabled:
        return 0
    if not HAS_HIVEX:
        logger.info("python3-hivex or python3-pefile not found. If you need Automatic Windows Installation support, "
                    "please install.")
        return 0

    profiles = api.profiles()
    systems = api.systems()
    templ = templar.Templar(api._collection_mgr)
    tgen = tftpgen.TFTPGen(api._collection_mgr)

    with open(os.path.join(settings.windows_template_dir, post_inst_cmd_template_name)) as template_win:
        post_tmpl_data = template_win.read()

    with open(os.path.join(settings.windows_template_dir, answerfile_template_name)) as template_win:
        tmpl_data = template_win.read()

    with open(os.path.join(settings.windows_template_dir, startnet_template_name)) as template_start:
        tmplstart_data = template_start.read()

    def gen_win_files(distro, meta):
        (kernel_path, kernel_name) = os.path.split(distro.kernel)
        distro_path = utils.find_distro_path(settings, distro)
        distro_dir = wim_file_name = os.path.join(settings.tftpboot_location, "images", distro.name)
        web_dir = os.path.join(settings.webdir, "images", distro.name)
        is_winpe = "winpe" in meta and meta['winpe'] != ""
        is_bcd = "bcd" in meta and meta['bcd'] != ""

        if "kernel" in meta:
            kernel_name = meta["kernel"]

        kernel_name = os.path.basename(kernel_name)
        is_wimboot = "wimboot" in kernel_name

        if is_wimboot:
            distro_path = os.path.join(settings.webdir, "distro_mirror", distro.name)
            kernel_path = os.path.join(distro_path, "boot")

            if "kernel" in meta and "wimboot" not in distro.kernel:
                tgen.copy_single_distro_file(os.path.join(settings.tftpboot_location, kernel_name), distro_dir, False)
                tgen.copy_single_distro_file(os.path.join(distro_dir, kernel_name), web_dir, True)

        if "post_install_script" in meta:
            post_install_dir = distro_path

            if distro.os_version not in ("XP", "2003"):
                post_install_dir = os.path.join(post_install_dir, "sources")

            post_install_dir = os.path.join(post_install_dir, "$OEM$", "$1")

            if not os.path.exists(post_install_dir):
                utils.mkdir(post_install_dir)

            data = templ.render(post_tmpl_data, meta, None)
            post_install_script = os.path.join(post_install_dir, meta["post_install_script"])
            logger.info('Build post install script: ' + post_install_script)
            with open(post_install_script, "w+") as pi_file:
                pi_file.write(data)

        if "answerfile" in meta:
            data = templ.render(tmpl_data, meta, None)
            answerfile_name = os.path.join(distro_dir, meta["answerfile"])
            logger.info('Build answer file: ' + answerfile_name)
            with open(answerfile_name, "w+") as answerfile:
                answerfile.write(data)
            tgen.copy_single_distro_file(answerfile_name, distro_path, False)
            tgen.copy_single_distro_file(answerfile_name, web_dir, True)

        if "kernel" in meta and "bootmgr" in meta:
            wk_file_name = os.path.join(distro_dir, kernel_name)
            wl_file_name = os.path.join(distro_dir, meta["bootmgr"])
            tl_file_name = os.path.join(kernel_path, "bootmgr.exe")

            if distro.os_version in ("XP", "2003") and not is_winpe:
                tl_file_name = os.path.join(kernel_path, "setupldr.exe")

                if len(meta["bootmgr"]) != 5:
                    logger.error("The loader name should be EXACTLY 5 character")
                    return 1

                pat1 = re.compile(br'NTLDR', re.IGNORECASE)
                pat2 = re.compile(br'winnt\.sif', re.IGNORECASE)
                with open(tl_file_name, 'rb') as file:
                    out = data = file.read()

                if "answerfile" in meta:
                    if len(meta["answerfile"]) != 9:
                        logger.error("The response file name should be EXACTLY 9 character")
                        return 1

                    out = pat2.sub(bytes(meta["answerfile"], 'utf-8'), data)
            else:
                if len(meta["bootmgr"]) != 11:
                    logger.error("The Boot manager file name should be EXACTLY 11 character")
                    return 1

                bcd_name = "bcd"
                if is_bcd:
                    bcd_name = meta["bcd"]
                    if len(bcd_name) != 3:
                        logger.error("The BCD file name should be EXACTLY 3 character")
                        return 1

                if not os.path.isfile(tl_file_name):
                    logger.error("File not found: %s" % tl_file_name)
                    return 1

                pat1 = re.compile(br'bootmgr\.exe', re.IGNORECASE)
                pat2 = re.compile(br'(\\.B.o.o.t.\\.)(B)(.)(C)(.)(D)', re.IGNORECASE)

                bcd_name = bytes("\\g<1>" + bcd_name[0] + "\\g<3>" + bcd_name[1] + "\\g<5>" + bcd_name[2], 'utf-8')
                with open(tl_file_name, 'rb') as file:
                    out = file.read()

                if not is_wimboot:
                    logger.info('Patching build Loader: %s' % wl_file_name)
                    out = pat2.sub(bcd_name, out)

            if tl_file_name != wl_file_name:
                logger.info('Build Loader: %s from %s' % (wl_file_name, tl_file_name))
                with open(wl_file_name, 'wb+') as file:
                    file.write(out)
                tgen.copy_single_distro_file(wl_file_name, web_dir, True)

            if not is_wimboot:
                if distro.os_version not in ("XP", "2003") or is_winpe:
                    pe = pefile.PE(wl_file_name, fast_load=True)
                    pe.OPTIONAL_HEADER.CheckSum = pe.generate_checksum()
                    pe.write(filename=wl_file_name)

                with open(distro.kernel, 'rb') as file:
                    data = file.read()
                out = pat1.sub(bytes(meta["bootmgr"], 'utf-8'), data)

                if wk_file_name != distro.kernel:
                    logger.info("Build PXEBoot: %s from %s" % (wk_file_name, distro.kernel))
                    with open(wk_file_name, 'wb+') as file:
                        file.write(out)
                    tgen.copy_single_distro_file(wk_file_name, web_dir, True)

        if is_bcd:
            obcd_file_name = os.path.join(kernel_path, "bcd")
            bcd_file_name = os.path.join(distro_dir, meta["bcd"])
            wim_file_name = 'winpe.wim'

            if not os.path.isfile(obcd_file_name):
                logger.error("File not found: %s" % obcd_file_name)
                return 1

            if is_winpe:
                wim_file_name = meta["winpe"]

            if is_wimboot:
                wim_file_name = '\\Boot\\' + wim_file_name
                sdi_file_name = '\\Boot\\' + 'boot.sdi'
            else:
                wim_file_name = os.path.join("/images", distro.name, wim_file_name)
                sdi_file_name = os.path.join("/images", distro.name, os.path.basename(distro.initrd))

            logger.info('Build BCD: %s from %s for %s' % (bcd_file_name, obcd_file_name, wim_file_name))
            bcdedit(obcd_file_name, bcd_file_name, wim_file_name, sdi_file_name)
            tgen.copy_single_distro_file(bcd_file_name, web_dir, True)

        if is_winpe:
            ps_file_name = os.path.join(distro_dir, meta["winpe"])
            wim_pl_name = os.path.join(kernel_path, "winpe.wim")

            cmd = ["/usr/bin/cp", "--reflink=auto", wim_pl_name, ps_file_name]
            utils.subprocess_call(cmd, shell=False)
            tgen.copy_single_distro_file(ps_file_name, web_dir, True)

            if os.path.exists(wimupdate):
                data = templ.render(tmplstart_data, meta, None)
                pi_file = tempfile.NamedTemporaryFile()
                pi_file.write(bytes(data, 'utf-8'))
                pi_file.flush()
                cmd = ["/usr/bin/wimdir %s 1 | /usr/bin/grep -i '^/Windows/System32/startnet.cmd$'" % ps_file_name]
                startnet_path = utils.subprocess_get(cmd, shell=True)[0:-1]
                cmd = [wimupdate, ps_file_name, "--command=add %s %s" % (pi_file.name, startnet_path)]
                utils.subprocess_call(cmd, shell=False)
                pi_file.close()

    for profile in profiles:
        distro = profile.get_conceptual_parent()

        if distro and distro.breed == "windows":
            logger.info('Profile: ' + profile.name)
            meta = utils.blender(api, False, profile)
            autoinstall_meta = meta.get("autoinstall_meta", {})
            meta.update(autoinstall_meta)
            gen_win_files(distro, meta)

    for system in systems:
        profile = system.get_conceptual_parent()
        autoinstall_meta = system.autoinstall_meta

        if not profile or not autoinstall_meta or autoinstall_meta == {}:
            continue

        distro = profile.get_conceptual_parent()

        if distro and distro.breed == "windows":
            logger.info('System: ' + system.name)
            meta = utils.blender(api, False, system)
            gen_win_files(distro, autoinstall_meta)
    return 0



"""

Copyright 2009, Red Hat, Inc and Others
Bill Peck <bpeck@redhat.com>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""

import glob
import os
import os.path
import logging


class LogTool:
    """
    Helpers for dealing with System logs, anamon, etc..
    """

    def __init__(self, collection_mgr, system, api):
        """
        Log library constructor requires a Cobbler system object.
        """
        self.system = system
        self.collection_mgr = collection_mgr
        self.settings = collection_mgr.settings()
        self.api = api
        self.logger = logging.getLogger()

    def clear(self):
        """
        Clears the system logs
        """
        anamon_dir = '/var/log/cobbler/anamon/%s' % self.system.name
        if os.path.isdir(anamon_dir):
            logs = list(filter(os.path.isfile, glob.glob('%s/*' % anamon_dir)))
        else:
            logs = []
            logging.info("No log-files found to delete for system: %s", self.system.name)

        for log in logs:
            try:
                with open(log, 'w') as f:
                    f.truncate()
            except IOError as e:
                self.logger.info("Failed to Truncate '%s':%s " % (log, e))
            except OSError as e:
                self.logger.info("Failed to Truncate '%s':%s " % (log, e))



"""
Copyright 2008-2009, Red Hat, Inc and Others
Michael DeHaan <michael.dehaan AT gmail>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""

from cobbler.cobbler_collections import collection
from cobbler.items import system as system
from cobbler import utils
from cobbler.cexceptions import CX


class Systems(collection.Collection):
    """
    Systems are hostnames/MACs/IP names and the associated profile
    they belong to.
    """

    @staticmethod
    def collection_type() -> str:
        return "system"

    @staticmethod
    def collection_types() -> str:
        return "systems"

    def factory_produce(self, api, item_dict):
        """
        Return a Distro forged from item_dict

        :param api: TODO
        :param item_dict: TODO
        :returns: TODO
        """
        new_system = system.System(api)
        new_system.from_dict(item_dict)
        return new_system

    def remove(self, name: str, with_delete: bool = True, with_sync: bool = True, with_triggers: bool = True,
               recursive: bool = False):
        """
        Remove element named 'name' from the collection

        :raises CX: In case the name of the object was not given.
        """
        name = name.lower()
        obj = self.find(name=name)

        if obj is None:
            raise CX("cannot delete an object that does not exist: %s" % name)

        if with_delete:
            if with_triggers:
                utils.run_triggers(self.api, obj, "/var/lib/cobbler/triggers/delete/system/pre/*", [])
            if with_sync:
                lite_sync = self.api.get_sync()
                lite_sync.remove_single_system(name)

        if obj.parent is not None and obj.name in obj.parent.children:
            obj.parent.children.remove(obj.name)
            self.api.serialize()

        self.lock.acquire()
        try:
            del self.listing[name]
        finally:
            self.lock.release()
        self.collection_mgr.serialize_delete(self, obj)
        if with_delete:
            if with_triggers:
                utils.run_triggers(self.api, obj, "/var/lib/cobbler/triggers/delete/system/post/*", [])
                utils.run_triggers(self.api, obj, "/var/lib/cobbler/triggers/change/*", [])



import pytest

from cobbler.modules.authentication import configfile


@pytest.fixture(scope="function")
def adjust_hashfunction():
    def _adjust_hashfunction(hashfunction):
        modulesconf = "/etc/cobbler/modules.conf"
        with open(modulesconf, 'r') as file:
            data = file.readlines()

        print(data[6])
        data[6] = "hash_algorithm = %s\n" % hashfunction
        print(data[6])

        with open(modulesconf, 'w') as file:
            file.writelines(data)
    return _adjust_hashfunction


class TestConfigfile:

    @pytest.mark.parametrize("hashfunction,test_input,exepcted_output", [
        ("sha3_512", "testtext", "eb17eb7a79798b31b3e625f2ff7a5cd05932254ca5f686764e9655274dde03c28ed4a7ab70b0637b5dc97e61da2ee07cc80e0c4f7d00feceb2d74cbe3a579698")
    ])
    def test_hashfun_positive(self, adjust_hashfunction, hashfunction, test_input, exepcted_output):
        # Arrange
        adjust_hashfunction(hashfunction)

        # Act
        result = configfile.hashfun(test_input)

        # Assert
        assert result == exepcted_output

    @pytest.mark.parametrize("hashfunction,test_input,exepcted_output", [
        ("md5", "testtext", "")
    ])
    def test_hashfun_negative(self, adjust_hashfunction, hashfunction, test_input, exepcted_output):
        # Arrange
        adjust_hashfunction(hashfunction)

        # Act & Assert
        with pytest.raises(ValueError):
            configfile.hashfun(test_input)

    def test_register(self):
        assert configfile.register() is "authn"

    @pytest.mark.parametrize("hashfunction, username, password", [
        ("md5", "cobbler", "cobbler")
    ])
    def test_authenticate_negative(self, adjust_hashfunction, hashfunction, username, password):
        # Arrange
        adjust_hashfunction(hashfunction)

        # Act & Assert
        with pytest.raises(ValueError):
            configfile.authenticate(None, username, password)

    @pytest.mark.parametrize("hashfunction, username, password", [
        ("sha3_512", "cobbler", "cobbler")
    ])
    def test_authenticate_positive(self, adjust_hashfunction, hashfunction, username, password):
        # Arrange
        adjust_hashfunction(hashfunction)

        # Act
        result = configfile.authenticate(None, username, password)

        # Assert
        assert result



import pytest

from cobbler import enums, utils, validate
from cobbler.api import CobblerAPI
from tests.conftest import does_not_raise


@pytest.mark.parametrize("test_architecture,test_raise", [
    (enums.Archs.X86_64, does_not_raise()),
    ("x86_64", does_not_raise()),
    ("abc", pytest.raises(ValueError)),
    (0, pytest.raises(TypeError))
])
def test_validate_arch(test_architecture, test_raise):
    # Arrange

    # Act
    with test_raise:
        result = validate.validate_arch(test_architecture)

        # Assert
        if isinstance(test_architecture, str):
            assert result.value == test_architecture
        elif isinstance(test_architecture, enums.Archs):
            assert result == test_architecture
        else:
            raise TypeError("result had a non expected result")


def test_validate_os_version():
    # Arrange
    utils.load_signatures("/var/lib/cobbler/distro_signatures.json")

    # Act
    result = validate.validate_os_version("rhel4", "redhat")

    # Assert
    assert result == "rhel4"


def test_validate_breed():
    # Arrange
    utils.load_signatures("/var/lib/cobbler/distro_signatures.json")

    # Act
    result = validate.validate_breed("redhat")

    # Assert
    assert result == "redhat"


def test_set_repos():
    # Arrange
    test_api = CobblerAPI()

    # Act
    # TODO: Test this also with the bypass check
    result = validate.validate_repos("testrepo1 testrepo2", test_api, bypass_check=True)

    # Assert
    assert result == ["testrepo1", "testrepo2"]


def test_set_virt_file_size():
    # Arrange

    # Act
    # TODO: Test multiple disks via comma separation
    result = validate.validate_virt_file_size("8")

    # Assert
    assert isinstance(result, float)
    assert result == 8


@pytest.mark.parametrize("test_driver,test_raise", [
    (enums.VirtDiskDrivers.RAW, does_not_raise()),
    (enums.VALUE_INHERITED, does_not_raise()),
    (enums.VirtDiskDrivers.INHERTIED, does_not_raise()),
    ("qcow2", does_not_raise()),
    ("bad_driver", pytest.raises(ValueError)),
    (0, pytest.raises(TypeError))
])
def test_set_virt_disk_driver(test_driver, test_raise):
    # Arrange

    # Act
    with test_raise:
        result = validate.validate_virt_disk_driver(test_driver)

        # Assert
        if isinstance(test_driver, str):
            assert result.value == test_driver
        elif isinstance(test_driver, enums.VirtDiskDrivers):
            assert result == test_driver
        else:
            raise TypeError("Unexpected type for value!")


@pytest.mark.parametrize("test_autoboot,expectation", [
    (True, does_not_raise()),
    (False, does_not_raise()),
    (0, does_not_raise()),
    (1, does_not_raise()),
    (2, does_not_raise()),
    ("Test", does_not_raise())
])
def test_set_virt_auto_boot(test_autoboot, expectation):
    # Arrange

    # Act
    with expectation:
        result = validate.validate_virt_auto_boot(test_autoboot)

        # Assert
        assert isinstance(result, bool)
        assert result is True or result is False


@pytest.mark.parametrize("test_input,expected_exception", [
    (True, does_not_raise()),
    (False, does_not_raise()),
    (0, does_not_raise()),
    (1, does_not_raise()),
    (5, does_not_raise()),
    ("", does_not_raise())
])
def test_set_virt_pxe_boot(test_input, expected_exception):
    # Arrange

    # Act
    with expected_exception:
        result = validate.validate_virt_pxe_boot(test_input)

        # Assert
        assert isinstance(result, bool)
        assert result or not result


def test_set_virt_ram():
    # Arrange

    # Act
    result = validate.validate_virt_ram(1024)

    # Assert
    assert result == 1024


@pytest.mark.parametrize("value,expected_exception", [
    ("qemu", does_not_raise()),
    (enums.VirtType.QEMU, does_not_raise()),
    (0, pytest.raises(TypeError))
])
def test_set_virt_type(value, expected_exception):
    # Arrange

    # Act
    with expected_exception:
        result = validate.validate_virt_type("qemu")

        # Assert
        if isinstance(value, str):
            assert result.value == value
        elif isinstance(value, enums.VirtType):
            assert result == value
        else:
            raise TypeError("Unexpected type for value!")


def test_set_virt_bridge():
    # Arrange

    # Act
    result = validate.validate_virt_bridge("testbridge")

    # Assert
    assert result == "testbridge"


def test_validate_virt_path():
    # Arrange
    test_location = "/somerandomfakelocation"

    # Act
    result = validate.validate_virt_path(test_location)

    # Assert
    assert result == test_location


@pytest.mark.parametrize("value,expected_exception", [
    (0, does_not_raise()),
    (5, does_not_raise()),
    (enums.VALUE_INHERITED, does_not_raise()),
    (False, does_not_raise()),
    (0.0, pytest.raises(TypeError)),
    (-5, pytest.raises(ValueError)),
    ("test", pytest.raises(TypeError))
])
def test_set_virt_cpus(value, expected_exception):
    # Arrange

    # Act
    with expected_exception:
        result = validate.validate_virt_cpus(value)

        # Assert
        if value == enums.VALUE_INHERITED:
            assert result == 0
        else:
            assert result == int(value)


def test_set_serial_device():
    # Arrange

    # Act
    result = validate.validate_serial_device(0)

    # Assert
    assert result == 0


def test_set_serial_baud_rate():
    # Arrange

    # Act
    result = validate.validate_serial_baud_rate(9600)

    # Assert
    assert result == enums.BaudRates.B9600


@pytest.mark.parametrize("test_value,expected_result", [
    ("test", False),
    (0, False),
    ("ftp://test/test", False),
    ("http://test_invalid/test", False),
    ("http://test§invalid/test", False),
    ("http://test.local/test", True),
    # ("http://test.local:80/test", True),
    ("http://test/test", True),
    ("http://@@server@@/test", True),
    ("http://10.0.0.1/test", True),
    ("http://fe80::989c:95ff:fe42:47bf/test", True),
])
def test_validate_boot_remote_file(test_value, expected_result):
    # Arrange

    # Act
    result = validate.validate_boot_remote_file(test_value)

    # Assert
    assert expected_result == result


@pytest.mark.parametrize("test_value,expected_result", [
    ("test", False),
    (0, False),
    ("http://test.local/test", False),
    ("http://test/test", False),
    ("ftp://test/test", False),
    # ("(tftp,10.0.0.1:invalid)/test", False),
    # ("(tftp,local_invalid)/test", False),
    ("(http,10.0.0.1)/test", True),
    ("(tftp,10.0.0.1)/test", True),
    ("(tftp,test.local)/test", True),
    ("(tftp,10.0.0.1:8080)/test", True)
])
def test_validate_grub_remote_file(test_value, expected_result):
    # Arrange

    # Act
    result = validate.validate_grub_remote_file(test_value)

    # Assert
    assert expected_result == result



import logging
from typing import Optional

import netaddr


def parse_grub_remote_file(file_location: str) -> Optional[str]:
    """
    Parses a URI which grub would try to load from the network.

    :param file_location: The location which grub would try to load from the network.
    :return: In case the URL could be parsed it is returned in the converted format. Otherwise None is returned.
    :raises TypeError: In case file_location is not of type ``str``.
    :raises ValueError: In case the file location does not contain a valid IPv4 or IPv6 address
    """
    if not isinstance(file_location, str):
        raise TypeError("\"file_location\" should be of type \"str\"")
    if file_location.startswith("ftp://"):
        logging.warning("FTP protocol not supported by GRUB. Only HTTP and TFTP [%s]", file_location)
        return None
    elif file_location.startswith("http://"):
        (server, delim, path) = file_location[7:].partition('/')
        prot = "http"
    elif file_location.startswith("tftp://"):
        (server, delim, path) = file_location[7:].partition('/')
        prot = "tftp"
    else:
        logging.warning("Unknown or unsupported protocol set for GRUB [%s]", file_location)
        return None

    if not (server.startswith("@@") and server.endswith("server@@")):
        if not (netaddr.valid_ipv4(server) or netaddr.valid_ipv6(server)):
            raise ValueError("Invalid remote file format %s\n%s is not a valid IP address" % (file_location, server))

    res = '(%s,%s)/%s' % (prot, server, path)
    # logging.info("Found remote grub file. Converted [%s] to [%s]", file_location, res)
    return res



"""
Copyright 2006-2009, Red Hat, Inc and Others
Michael DeHaan <michael.dehaan AT gmail>
John Eckersberg <jeckersb@redhat.com>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""

from typing import Dict, List, Callable, Any, Optional

import glob
import gzip
import os
import os.path
import re
import shutil
import stat

import magic

HAS_HIVEX = True
try:
    import hivex
    from hivex.hive_types import REG_SZ
except Exception:
    HAS_HIVEX = False

from cobbler.items import profile, distro
from cobbler.cexceptions import CX
from cobbler import enums, utils
from cobbler.manager import ManagerModule

import cobbler.items.repo as item_repo

# Import aptsources module if available to obtain repo mirror.
try:
    from aptsources import distro as debdistro
    from aptsources import sourceslist

    apt_available = True
except:
    apt_available = False

MANAGER = None


def register() -> str:
    """
    The mandatory Cobbler module registration hook.
    """
    return "manage/import"


def import_walker(top: str, func: Callable, arg: Any):
    """
    Directory tree walk with callback function.

    For each directory in the directory tree rooted at top (including top itself, but excluding '.' and '..'), call
    ``func(arg, dirname, fnames)``. dirname is the name of the directory, and fnames a list of the names of the files
    and subdirectories in dirname (excluding '.' and '..').  ``func`` may modify the ``fnames`` list in-place (e.g. via
    ``del`` or ``slice`` assignment), and walk will only recurse into the subdirectories whose names remain in
    ``fnames``; this can be used to implement a filter, or to impose a specific order of visiting. No semantics are
    defined for, or required of, ``arg``, beyond that arg is always passed to ``func``. It can be used, e.g., to pass
    a filename pattern, or a mutable object designed to accumulate statistics.

    :param top: The most top directory for which func should be run.
    :param func: A function which is called as described in the above description.
    :param arg: Passing ``None`` for this is common.
    """
    try:
        names = os.listdir(top)
    except os.error:
        return
    func(arg, top, names)
    for name in names:
        name = os.path.join(top, name)
        try:
            st = os.lstat(name)
        except os.error:
            continue
        if stat.S_ISDIR(st.st_mode):
            import_walker(name, func, arg)


class _ImportSignatureManager(ManagerModule):

    @staticmethod
    def what() -> str:
        """
        Identifies what service this manages.

        :return: Always will return ``import/signatures``.
        """
        return "import/signatures"

    def __init__(self, collection_mgr):
        super().__init__(collection_mgr)

        self.signature = None
        self.found_repos = {}

    def get_file_lines(self, filename: str):
        """
        Get lines from a file, which may or may not be compressed. If compressed then it will be uncompressed using
        ``gzip`` as the algorithm.

        :param filename: The name of the file to be read.
        :return: An array with all the lines.
        """
        ftype = magic.detect_from_filename(filename)
        if ftype.mime_type == "application/gzip":
            try:
                with gzip.open(filename, 'r') as f:
                    return f.readlines()
            except:
                pass
        if ftype.mime_type == "application/x-ms-wim":
            cmd = "/usr/bin/wiminfo"
            if os.path.exists(cmd):
                cmd = "%s %s" % (cmd, filename)
                return utils.subprocess_get(cmd).splitlines()

            self.logger.info("no %s found, please install wimlib-utils", cmd)
        elif ftype.mime_type == "text/plain":
            with open(filename, 'r') as f:
                return f.readlines()
        else:
            self.logger.info('Could not detect the filetype and read the content of file "%s". Returning nothing.',
                             filename)
        return []

    def run(self, path: str, name: str, network_root=None, autoinstall_file=None, arch: Optional[str] = None,
            breed=None, os_version=None):
        """
        This is the main entry point in a manager. It is a required function for import modules.

        :param path: the directory we are scanning for files
        :param name: the base name of the distro
        :param network_root: the remote path (nfs/http/ftp) for the distro files
        :param autoinstall_file: user-specified response file, which will override the default
        :param arch: user-specified architecture
        :param breed: user-specified breed
        :param os_version: user-specified OS version
        :raises CX
        """
        self.name = name
        self.network_root = network_root
        self.autoinstall_file = autoinstall_file
        self.arch = arch
        self.breed = breed
        self.os_version = os_version

        self.path = path
        self.rootdir = path
        self.pkgdir = path

        # some fixups for the XMLRPC interface, which does not use "None"
        if self.arch == "":
            self.arch = None

        if self.name == "":
            self.name = None

        if self.autoinstall_file == "":
            self.autoinstall_file = None

        if self.os_version == "":
            self.os_version = None

        if self.network_root == "":
            self.network_root = None

        if self.os_version and not self.breed:
            utils.die("OS version can only be specified when a specific breed is selected")

        self.signature = self.scan_signatures()
        if not self.signature:
            error_msg = "No signature matched in %s" % path
            self.logger.error(error_msg)
            raise CX(error_msg)

        # now walk the filesystem looking for distributions that match certain patterns
        self.logger.info("Adding distros from path %s:" % self.path)
        distros_added = []
        import_walker(self.path, self.distro_adder, distros_added)

        if len(distros_added) == 0:
            if self.breed == "windows":
                cmd_path = "/usr/bin/wimexport"
                bootwim_path = os.path.join(self.path, "sources", "boot.wim")
                dest_path = os.path.join(self.path, "boot")
                if os.path.exists(cmd_path) and os.path.exists(bootwim_path):
                    winpe_path = os.path.join(dest_path, "winpe.wim")
                    if not os.path.exists(dest_path):
                        utils.mkdir(dest_path)
                    rc = utils.subprocess_call([cmd_path, bootwim_path, "1",
                                               winpe_path, "--boot"], shell=False)
                    if rc == 0:
                        cmd = ["/usr/bin/wimdir %s 1 | /usr/bin/grep -i '^/Windows/Boot/PXE$'" % winpe_path]
                        pxe_path = utils.subprocess_get(cmd, shell=True)[0:-1]
                        cmd = ["/usr/bin/wimdir %s 1 | /usr/bin/grep -i '^/Windows/System32/config/SOFTWARE$'"
                               % winpe_path]
                        config_path = utils.subprocess_get(cmd, shell=True)[0:-1]
                        cmd_path = "/usr/bin/wimextract"
                        rc = utils.subprocess_call([cmd_path, bootwim_path, "1",
                                                    "%s/pxeboot.n12" % pxe_path,
                                                    "%s/bootmgr.exe" % pxe_path,
                                                    config_path,
                                                    "--dest-dir=%s" % dest_path,
                                                    "--no-acls", "--no-attributes"], shell=False)
                        if rc == 0:
                            if HAS_HIVEX:
                                software = os.path.join(dest_path, os.path.basename(config_path))
                                h = hivex.Hivex(software, write=True)
                                root = h.root()
                                node = h.node_get_child(root, "Microsoft")
                                node = h.node_get_child(node, "Windows NT")
                                node = h.node_get_child(node, "CurrentVersion")
                                h.node_set_value(node, {"key": "SystemRoot", "t": REG_SZ,
                                                        "value": "x:\\Windows\0".encode(encoding="utf-16le")})
                                node = h.node_get_child(node, "WinPE")

                                # remove the key InstRoot from the registry
                                values = h.node_values(node)
                                new_values = []

                                for value in values:
                                    keyname = h.value_key(value)

                                    if keyname == "InstRoot":
                                        continue

                                    val = h.node_get_value(node, keyname)
                                    valtype = h.value_type(val)[0]
                                    value2 = h.value_value(val)[1]
                                    valobject = {"key": keyname, "t": int(valtype), "value": value2}
                                    new_values.append(valobject)

                                h.node_set_values(node, new_values)
                                h.commit(software)

                                cmd_path = "/usr/bin/wimupdate"
                                rc = utils.subprocess_call([cmd_path, winpe_path, "--command=add %s %s"
                                                           % (software, config_path)], shell=False)
                                os.remove(software)
                            else:
                                self.logger.info("python3-hivex not found. If you need Automatic Windows "
                                                 "Installation support, please install.")
                            import_walker(self.path, self.distro_adder, distros_added)

        if len(distros_added) == 0:
            self.logger.warning("No distros imported, bailing out")
            return

        # find out if we can auto-create any repository records from the install tree
        if self.network_root is None:
            self.logger.info("associating repos")
            # FIXME: this automagic is not possible (yet) without mirroring
            self.repo_finder(distros_added)

    def scan_signatures(self):
        """
        Loop through the signatures, looking for a match for both the signature directory and the version file.
        """
        sigdata = self.api.get_signatures()
        # self.logger.debug("signature cache: %s" % str(sigdata))
        for breed in list(sigdata["breeds"].keys()):
            if self.breed and self.breed != breed:
                continue
            for version in list(sigdata["breeds"][breed].keys()):
                if self.os_version and self.os_version != version:
                    continue
                for sig in sigdata["breeds"][breed][version]["signatures"]:
                    pkgdir = os.path.join(self.path, sig)
                    if os.path.exists(pkgdir):
                        self.logger.debug("Found a candidate signature: breed=%s, version=%s" % (breed, version))
                        f_re = re.compile(sigdata["breeds"][breed][version]["version_file"])
                        for (root, subdir, fnames) in os.walk(self.path):
                            for fname in fnames + subdir:
                                if f_re.match(fname):
                                    # if the version file regex exists, we use it to scan the contents of the target
                                    # version file to ensure it's the right version
                                    if sigdata["breeds"][breed][version]["version_file_regex"]:
                                        vf_re = re.compile(sigdata["breeds"][breed][version]["version_file_regex"])
                                        vf_lines = self.get_file_lines(os.path.join(root, fname))
                                        for line in vf_lines:
                                            if vf_re.match(line):
                                                self.logger.debug("Found a matching signature: breed=%s, version=%s",
                                                                  breed, version)
                                                if not self.breed:
                                                    self.breed = breed
                                                if not self.os_version:
                                                    self.os_version = version
                                                if not self.autoinstall_file:
                                                    self.autoinstall_file = sigdata["breeds"][breed][version]["default_autoinstall"]
                                                self.pkgdir = pkgdir
                                                return sigdata["breeds"][breed][version]
                        break
        return None

    # required function for import modules
    def get_valid_arches(self) -> list:
        """
        Get all valid architectures from the signature file.

        :return: An empty list or all valid architectures.
        """
        if self.signature:
            return sorted(self.signature["supported_arches"], key=lambda s: -1 * len(s))
        return []

    def get_valid_repo_breeds(self) -> list:
        """
        Get all valid repository architectures from the signatures file.

        :return: An empty list or all valid architectures.
        """
        if self.signature:
            return self.signature["supported_repo_breeds"]
        return []

    def distro_adder(self, distros_added, dirname: str, fnames):
        """
        This is an import_walker routine that finds distributions in the directory to be scanned and then creates them.

        :param distros_added: Unknown what this currently does.
        :param dirname: Unknown what this currently does.
        :param fnames: Unknown what this currently does.
        """

        re_krn = re.compile(self.signature["kernel_file"])
        re_img = re.compile(self.signature["initrd_file"])

        # make sure we don't mismatch PAE and non-PAE types
        initrd = None
        kernel = None
        pae_initrd = None
        pae_kernel = None

        for x in fnames:
            adtls = []

            # Most of the time we just want to ignore isolinux directories, unless this is one of the oddball distros
            # where we do want it.
            if dirname.find("isolinux") != -1 and not self.signature["isolinux_ok"]:
                continue

            fullname = os.path.join(dirname, x)
            if os.path.islink(fullname) and os.path.isdir(fullname):
                if fullname.startswith(self.path):
                    # Prevent infinite loop with Sci Linux 5
                    # self.logger.warning("avoiding symlink loop")
                    continue
                self.logger.info("following symlink: %s" % fullname)
                import_walker(fullname, self.distro_adder, distros_added)

            if re_img.match(x):
                if x.find("PAE") == -1:
                    initrd = os.path.join(dirname, x)
                else:
                    pae_initrd = os.path.join(dirname, x)

            if re_krn.match(x):
                if x.find("PAE") == -1:
                    kernel = os.path.join(dirname, x)
                else:
                    pae_kernel = os.path.join(dirname, x)

            # if we've collected a matching kernel and initrd pair, turn them in and add them to the list
            if initrd is not None and kernel is not None:
                adtls.append(self.add_entry(dirname, kernel, initrd))
                kernel = None
                initrd = None
            elif pae_initrd is not None and pae_kernel is not None:
                adtls.append(self.add_entry(dirname, pae_kernel, pae_initrd))
                pae_kernel = None
                pae_initrd = None

            for adtl in adtls:
                distros_added.extend(adtl)

    def add_entry(self, dirname: str, kernel, initrd):
        """
        When we find a directory with a valid kernel/initrd in it, create the distribution objects as appropriate and
        save them. This includes creating xen and rescue distros/profiles if possible.

        :param dirname: Unkown what this currently does.
        :param kernel: Unkown what this currently does.
        :param initrd: Unkown what this currently does.
        :return: Unkown what this currently does.
        """

        # build a proposed name based on the directory structure
        proposed_name = self.get_proposed_name(dirname, kernel)

        # build a list of arches found in the packages directory
        archs = self.learn_arch_from_tree()
        if not archs and self.arch:
            archs.append(self.arch)
        else:
            if self.arch and self.arch not in archs:
                utils.die("Given arch (%s) not found on imported tree %s" % (self.arch, self.path))

        if len(archs) == 0:
            self.logger.error("No arch could be detected in %s, and none was specified via the --arch option" % dirname)
            return []
        elif len(archs) > 1:
            self.logger.warning("- Warning : Multiple archs found : %s" % archs)

        distros_added = []
        for pxe_arch in archs:
            name = proposed_name + "-" + pxe_arch
            existing_distro = self.distros.find(name=name)

            if existing_distro is not None:
                self.logger.warning("skipping import, as distro name already exists: %s" % name)
                continue
            else:
                self.logger.info("creating new distro: %s" % name)
                new_distro = distro.Distro(self.api)

            if name.find("-autoboot") != -1:
                # this is an artifact of some EL-3 imports
                continue

            new_distro.name = name
            new_distro.kernel = kernel
            new_distro.initrd = initrd
            new_distro.arch = pxe_arch
            new_distro.breed = self.breed
            new_distro.os_version = self.os_version
            new_distro.kernel_options = self.signature.get("kernel_options", "")
            new_distro.kernel_options_post = self.signature.get("kernel_options_post", "")
            new_distro.template_files = self.signature.get("template_files", "")

            boot_files: Dict[str, str] = {}
            for boot_file in self.signature["boot_files"]:
                boot_files['$local_img_path/%s' % boot_file] = '%s/%s' % (self.path, boot_file)
            new_distro.boot_files = boot_files

            self.configure_tree_location(new_distro)

            self.distros.add(new_distro, save=True)
            distros_added.append(new_distro)

            # see if the profile name is already used, if so, skip it and
            # do not modify the existing profile

            existing_profile = self.profiles.find(name=name)

            if existing_profile is None:
                self.logger.info("creating new profile: %s" % name)
                new_profile = profile.Profile(self.api)
            else:
                self.logger.info("skipping existing profile, name already exists: %s" % name)
                continue

            new_profile.name = name
            new_profile.distro = name
            new_profile.autoinstall = self.autoinstall_file

            # depending on the name of the profile we can
            # define a good virt-type for usage with koan
            if name.find("-xen") != -1:
                new_profile.virt_type = enums.VirtType.XENPV
            elif name.find("vmware") != -1:
                new_profile.virt_type = enums.VirtType.VMWARE
            else:
                new_profile.virt_type = enums.VirtType.KVM

            if self.breed == "windows":
                dest_path = os.path.join(self.path, "boot")
                bootmgr_path = os.path.join(dest_path, "bootmgr.exe")
                bcd_path = os.path.join(dest_path, "bcd")
                winpe_path = os.path.join(dest_path, "winpe.wim")
                if os.path.exists(bootmgr_path) and os.path.exists(bcd_path) and os.path.exists(winpe_path):
                    new_profile.autoinstall_meta = {"kernel": os.path.basename(kernel),
                                                    "bootmgr": "bootmgr.exe",
                                                    "bcd": "bcd",
                                                    "winpe": "winpe.wim",
                                                    "answerfile": "autounattended.xml"}

            self.profiles.add(new_profile, save=True)

        return distros_added

    def learn_arch_from_tree(self) -> list:
        """
        If a distribution is imported from DVD, there is a good chance the path doesn't contain the arch and we should
        add it back in so that it's part of the meaningful name ... so this code helps figure out the arch name.  This
        is important for producing predictable distro names (and profile names) from differing import sources.

        :return: The guessed architecture from a distribution dvd.
        """

        result = {}

        # FIXME : this is called only once, should not be a walk
        import_walker(self.path, self.arch_walker, result)

        if result.pop("amd64", False):
            result["x86_64"] = 1
        if result.pop("i686", False):
            result["i386"] = 1
        if result.pop("i586", False):
            result["i386"] = 1
        if result.pop("x86", False):
            result["i386"] = 1

        return list(result.keys())

    def arch_walker(self, foo: dict, dirname: str, fnames: list):
        """
        Function for recursively searching through a directory for a kernel file matching a given architecture, called
        by ``learn_arch_from_tree()``

        :param foo: Into this dict there will be put additional meta information.
        :param dirname: The directory name where the kernel can be found.
        :param fnames: This should be a list like object which will be looped over.
        """

        re_krn = re.compile(self.signature["kernel_arch"])

        # try to find a kernel header RPM and then look at it's arch.
        for x in fnames:
            if re_krn.match(x):
                if self.signature["kernel_arch_regex"]:
                    re_krn2 = re.compile(self.signature["kernel_arch_regex"])
                    krn_lines = self.get_file_lines(os.path.join(dirname, x))
                    for line in krn_lines:
                        m = re_krn2.match(line)
                        if m:
                            for group in m.groups():
                                group = group.lower()
                                if group in self.get_valid_arches():
                                    foo[group] = 1
                else:
                    for arch in self.get_valid_arches():
                        if x.find(arch) != -1:
                            foo[arch] = 1
                            break
                    for arch in ["i686", "amd64"]:
                        if x.find(arch) != -1:
                            foo[arch] = 1
                            break

    def get_proposed_name(self, dirname: str, kernel=None) -> str:
        """
        Given a directory name where we have a kernel/initrd pair, try to autoname the distribution (and profile) object
        based on the contents of that path.

        :param dirname: The directory where the distribution is living in.
        :param kernel: The kernel of that distro.
        :return: The name which is recommended.
        """

        if self.network_root is not None:
            name = self.name
        else:
            # remove the part that says /var/www/cobbler/distro_mirror/name
            name = "-".join(dirname.split("/")[5:])

        if kernel is not None:
            if kernel.find("PAE") != -1 and name.find("PAE") == -1:
                name += "-PAE"
            if kernel.find("xen") != -1 and name.find("xen") == -1:
                name += "-xen"

        # Clear out some cruft from the proposed name
        name = name.replace("--", "-")
        for x in ("-netboot", "-ubuntu-installer", "-amd64", "-i386",
                  "-images", "-pxeboot", "-install", "-isolinux", "-boot", "-suseboot",
                  "-loader", "-os", "-tree", "var-www-cobbler-", "distro_mirror-"):
            name = name.replace(x, "")

        # remove any architecture name related string, as real arch will be appended later
        name = name.replace("chrp", "ppc64")
        for separator in ['-', '_', '.']:
            for arch in ["i386", "x86_64", "ia64", "ppc64le", "ppc64el", "ppc64", "ppc32", "ppc", "x86", "s390x",
                         "s390", "386", "amd"]:
                name = name.replace("%s%s" % (separator, arch), "")

        return name

    def configure_tree_location(self, distribution: distro.Distro):
        """
        Once a distribution is identified, find the part of the distribution that has the URL in it that we want to use
        for automating the Linux distribution installation, and create a autoinstall_meta variable $tree that contains
        this.

        :param distribution: The distribution object for that the tree should be configured.
        """

        base = self.rootdir

        # how we set the tree depends on whether an explicit network_root was specified
        if self.network_root is None:
            dest_link = os.path.join(self.settings.webdir, "links", distribution.name)
            # create the links directory only if we are mirroring because with SELinux Apache can't symlink to NFS
            # (without some doing)
            if not os.path.exists(dest_link):
                try:
                    self.logger.info("trying symlink: %s -> %s" % (str(base), str(dest_link)))
                    os.symlink(base, dest_link)
                except:
                    # FIXME: This shouldn't happen but I've seen it ... debug ...
                    self.logger.warning(
                        "symlink creation failed: %(base)s, %(dest)s" % {"base": base, "dest": dest_link})
            tree = "http://@@http_server@@/cblr/links/%s" % distribution.name
            self.set_install_tree(distribution, tree)
        else:
            # Where we assign the automated installation file source is relative to our current directory and the input
            # start directory in the crawl. We find the path segments between and tack them on the network source
            # path to find the explicit network path to the distro that Anaconda can digest.
            tail = utils.path_tail(self.path, base)
            tree = self.network_root[:-1] + tail
            self.set_install_tree(distribution, tree)

    def set_install_tree(self, distribution: distro.Distro, url: str):
        """
        Simple helper function to set the tree automated installation metavariable.

        :param distribution: The distribution object for which the install tree should be set.
        :param url: The url for the tree.
        """
        distribution.autoinstall_meta["tree"] = url

    # ==========================================================================
    # Repo Functions

    def repo_finder(self, distros_added: List[distro.Distro]):
        """
        This routine looks through all distributions and tries to find any applicable repositories in those
        distributions for post-install usage.

        :param distros_added: This is an iteratable set of distributions.
        """
        for repo_breed in self.get_valid_repo_breeds():
            self.logger.info("checking for %s repo(s)" % repo_breed)
            repo_adder = None
            if repo_breed == "yum":
                repo_adder = self.yum_repo_adder
            elif repo_breed == "rhn":
                repo_adder = self.rhn_repo_adder
            elif repo_breed == "rsync":
                repo_adder = self.rsync_repo_adder
            elif repo_breed == "apt":
                repo_adder = self.apt_repo_adder
            else:
                self.logger.warning("skipping unknown/unsupported repo breed: %s" % repo_breed)
                continue

            for current_distro_added in distros_added:
                if current_distro_added.kernel.find("distro_mirror") != -1:
                    repo_adder(current_distro_added)
                    self.distros.add(current_distro_added, save=True, with_triggers=False)
                else:
                    self.logger.info("skipping distro %s since it isn't mirrored locally" % current_distro_added.name)

    # ==========================================================================
    # yum-specific

    def yum_repo_adder(self, distro: distro.Distro):
        """
        For yum, we recursively scan the rootdir for repos to add

        :param distro: The distribution object to scan and possibly add.
        """
        self.logger.info("starting descent into %s for %s" % (self.rootdir, distro.name))
        import_walker(self.rootdir, self.yum_repo_scanner, distro)

    def yum_repo_scanner(self, distro: distro.Distro, dirname: str, fnames):
        """
        This is an import_walker routine that looks for potential yum repositories to be added to the configuration for
        post-install usage.

        :param distro: The distribution object to check for.
        :param dirname: The folder with repositories to check.
        :param fnames: Unkown what this does exactly.
        """

        matches = {}
        for x in fnames:
            if x == "base" or x == "repodata":
                self.logger.info("processing repo at : %s" % dirname)
                # only run the repo scanner on directories that contain a comps.xml
                gloob1 = glob.glob("%s/%s/*comps*.xml" % (dirname, x))
                if len(gloob1) >= 1:
                    if dirname in matches:
                        self.logger.info("looks like we've already scanned here: %s" % dirname)
                        continue
                    self.logger.info("need to process repo/comps: %s" % dirname)
                    self.yum_process_comps_file(dirname, distro)
                    matches[dirname] = 1
                else:
                    self.logger.info("directory %s is missing xml comps file, skipping" % dirname)
                    continue

    def yum_process_comps_file(self, comps_path: str, distribution: distro.Distro):
        """
        When importing Fedora/EL certain parts of the install tree can also be used as yum repos containing packages
        that might not yet be available via updates in yum. This code identifies those areas. Existing repodata will be
        used as-is, but repodate is created for earlier, non-yum based, installers.

        :param comps_path: Not know what this is exactly for.
        :param distribution: The distributions to check.
        """

        if os.path.exists(os.path.join(comps_path, "repodata")):
            keeprepodata = True
            masterdir = "repodata"
        else:
            # older distros...
            masterdir = "base"
            keeprepodata = False

        # figure out what our comps file is ...
        self.logger.info("looking for %(p1)s/%(p2)s/*comps*.xml" % {"p1": comps_path, "p2": masterdir})
        files = glob.glob("%s/%s/*comps*.xml" % (comps_path, masterdir))
        if len(files) == 0:
            self.logger.info("no comps found here: %s" % os.path.join(comps_path, masterdir))
            return  # no comps xml file found

        # pull the filename from the longer part
        comps_file = files[0].split("/")[-1]

        try:
            # Store the yum configs on the filesystem so we can use them later. And configure them in the automated
            # installation file post section, etc.

            counter = len(distribution.source_repos)

            # find path segment for yum_url (changing filesystem path to http:// trailing fragment)
            seg = comps_path.rfind("distro_mirror")
            urlseg = comps_path[(seg + len("distro_mirror") + 1):]

            fname = os.path.join(self.settings.webdir,
                                 "distro_mirror", "config", "%s-%s.repo" % (distribution.name, counter))

            repo_url = "http://@@http_server@@/cobbler/distro_mirror/config/%s-%s.repo" % (distribution.name, counter)
            repo_url2 = "http://@@http_server@@/cobbler/distro_mirror/%s" % urlseg

            distribution.source_repos.append([repo_url, repo_url2])

            config_dir = os.path.dirname(fname)
            if not os.path.exists(config_dir):
                os.makedirs(config_dir)

            # NOTE: the following file is now a Cheetah template, so it can be remapped during sync, that's why we have
            # the @@http_server@@ left as templating magic.
            # repo_url2 is actually no longer used. (?)

            with open(fname, "w+") as config_file:
                config_file.write("[core-%s]\n" % counter)
                config_file.write("name=core-%s\n" % counter)
                config_file.write("baseurl=http://@@http_server@@/cobbler/distro_mirror/%s\n" % urlseg)
                config_file.write("enabled=1\n")
                config_file.write("gpgcheck=0\n")
                config_file.write("priority=$yum_distro_priority\n")

            # Don't run creatrepo twice -- this can happen easily for Xen and PXE, when they'll share same repo files.
            if keeprepodata:
                self.logger.info("Keeping repodata as-is :%s/repodata" % comps_path)
                self.found_repos[comps_path] = 1

            elif comps_path not in self.found_repos:
                utils.remove_yum_olddata(comps_path)
                cmd = "createrepo %s --groupfile %s %s" % (
                    self.settings.createrepo_flags, os.path.join(comps_path, masterdir, comps_file), comps_path)
                utils.subprocess_call(cmd, shell=True)
                self.found_repos[comps_path] = 1
                # For older distros, if we have a "base" dir parallel with "repodata", we need to copy comps.xml up
                # one...
                p1 = os.path.join(comps_path, "repodata", "comps.xml")
                p2 = os.path.join(comps_path, "base", "comps.xml")
                if os.path.exists(p1) and os.path.exists(p2):
                    shutil.copyfile(p1, p2)
        except:
            self.logger.error("error launching createrepo (not installed?), ignoring")
            utils.log_exc()

    # ==========================================================================
    # apt-specific

    def apt_repo_adder(self, distribution: distro.Distro):
        """
        Automatically import apt repositories when importing signatures.

        :param distribution: The distribution to scan for apt repositories.
        """
        self.logger.info("adding apt repo for %s" % distribution.name)
        # Obtain repo mirror from APT if available
        mirror = ""
        if apt_available:
            # Example returned URL: http://us.archive.ubuntu.com/ubuntu
            mirror = self.get_repo_mirror_from_apt()
        if not mirror:
            mirror = "http://archive.ubuntu.com/ubuntu"

        repo = item_repo.Repo(self.collection_mgr)
        repo.breed = enums.RepoBreeds.APT
        repo.arch = distribution.arch
        repo.keep_updated = True
        repo.apt_components = "main universe"  # TODO: make a setting?
        repo.apt_dists = "%s %s-updates %s-security" % ((distribution.os_version,) * 3)
        repo.name = distribution.name
        repo.os_version = distribution.os_version

        if distribution.breed == "ubuntu":
            repo.mirror = mirror
        else:
            # NOTE : The location of the mirror should come from timezone
            repo.mirror = "http://ftp.%s.debian.org/debian/dists/%s" % ('us', distribution.os_version)

        self.logger.info("Added repos for %s" % distribution.name)
        repos = self.collection_mgr.repos()
        repos.add(repo, save=True)
        # FIXME: Add the found/generated repos to the profiles that were created during the import process

    def get_repo_mirror_from_apt(self):
        """
        This tries to determine the apt mirror/archive to use (when processing repos) if the host machine is Debian or
        Ubuntu.

        :return: False if the try fails or otherwise the mirrors.
        """
        try:
            sources = sourceslist.SourcesList()
            release = debdistro.get_distro()
            release.get_sources(sources)
            mirrors = release.get_server_list()
            for mirror in mirrors:
                if mirror[2]:
                    return mirror[1]
        except:
            return False

    # ==========================================================================
    # rhn-specific

    def rhn_repo_adder(self, distribution: distro.Distro):
        """
        Not currently used.

        :param distribution: Not used currently.
        """
        return

    # ==========================================================================
    # rsync-specific

    def rsync_repo_adder(self, distribution: distro.Distro):
        """
        Not currently used.

        :param distribution: Not used currently.
        """
        return


# ==========================================================================


def get_import_manager(collection_mgr):
    """
    Get an instance of the import manager which enables you to import various things.

    :param collection_mgr: The collection Manager instance of Cobbler
    :return: The object to import data with.
    """
    # Singleton used, therefore ignoring 'global'
    global MANAGER  # pylint: disable=global-statement

    if not MANAGER:
        MANAGER = _ImportSignatureManager(collection_mgr)
    return MANAGER



"""
Authorization module that allows everything, which is the default for new Cobbler installs.

Copyright 2007-2009, Red Hat, Inc and Others
Michael DeHaan <michael.dehaan AT gmail>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""


def register() -> str:
    """
    The mandatory Cobbler module registration hook.

    :return: Always "authz"
    """
    return "authz"


def authorize(api_handle, user, resource, arg1=None, arg2=None) -> bool:
    """
    Validate a user against a resource.
    NOTE: acls are not enforced as there is no group support in this module

    :param api_handle: This parameter is not used currently.
    :param user: This parameter is not used currently.
    :param resource: This parameter is not used currently.
    :param arg1: This parameter is not used currently.
    :param arg2: This parameter is not used currently.
    :return: Always True
    """
    return True



"""
Copyright 2006-2009, Red Hat, Inc and Others
Michael DeHaan <michael.dehaan AT gmail>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""


import glob
import logging
import os
import re

from cobbler import utils


class CobblerCheck:
    """
    Validates whether the system is reasonably well configured for
    serving up content. This is the code behind 'cobbler check'.
    """

    def __init__(self, collection_mgr):
        """
        Constructor

        :param collection_mgr: The collection manager which holds all information.
        """
        self.collection_mgr = collection_mgr
        self.settings = collection_mgr.settings()
        self.logger = logging.getLogger()
        self.checked_family = ""

    def run(self):
        """
        The CLI usage is "cobbler check" before "cobbler sync".

        :return: None if there are no errors, otherwise returns a list of things to correct prior to running application
                 'for real'.
        """
        status = []
        self.checked_family = utils.get_family()
        self.check_name(status)
        self.check_selinux(status)
        if self.settings.manage_dhcp:
            mode = self.collection_mgr.api.get_sync().dhcp.what()
            if mode == "isc":
                self.check_dhcpd_bin(status)
                self.check_dhcpd_conf(status)
                self.check_service(status, "dhcpd")
            elif mode == "dnsmasq":
                self.check_dnsmasq_bin(status)
                self.check_service(status, "dnsmasq")

        if self.settings.manage_dns:
            mode = self.collection_mgr.api.get_sync().dns.what()
            if mode == "bind":
                self.check_bind_bin(status)
                self.check_service(status, "named")
            elif mode == "dnsmasq" and not self.settings.manage_dhcp:
                self.check_dnsmasq_bin(status)
                self.check_service(status, "dnsmasq")

        mode = self.collection_mgr.api.get_sync().tftpd.what()
        if mode == "in_tftpd":
            self.check_tftpd_dir(status)
        elif mode == "tftpd_py":
            self.check_ctftpd_dir(status)

        self.check_service(status, "cobblerd")

        self.check_bootloaders(status)
        self.check_for_wget_curl(status)
        self.check_rsync_conf(status)
        self.check_iptables(status)
        self.check_yum(status)
        self.check_debmirror(status)
        self.check_for_ksvalidator(status)
        self.check_for_default_password(status)
        self.check_for_unreferenced_repos(status)
        self.check_for_unsynced_repos(status)
        self.check_for_cman(status)

        return status

    def check_for_ksvalidator(self, status):
        """
        Check if the ``ksvalidator`` is present in ``/usr/bin``.

        :param status: The status list with possible problems. The status list with possible problems.
        """
        # FIXME: This tools is cross-platform via Python. Thus all distros can have it.
        # ubuntu also identifies as "debian"
        if self.checked_family in ["debian", "suse"]:
            return

        if not os.path.exists("/usr/bin/ksvalidator"):
            status.append("ksvalidator was not found, install pykickstart")

    def check_for_cman(self, status):
        """
        Check if the fence agents are available. This is done through checking if the binary ``fence_ilo`` is present
        in ``/sbin`` or ``/usr/sbin``.

        :param status: The status list with possible problems. The status list with possible problems.
        """
        # not doing rpm -q here to be cross-distro friendly
        if not os.path.exists("/sbin/fence_ilo") and not os.path.exists("/usr/sbin/fence_ilo"):
            status.append("fencing tools were not found, and are required to use the (optional) power management "
                          "features. install cman or fence-agents to use them")

    def check_service(self, status, which, notes=""):
        """
        Check if the service command is available or the old init.d system has to be used.

        :param status: The status list with possible problems.
        :param which: The service to check for.
        :param notes: A manual not to attach.
        """
        # TODO: This should be improved. Removing would be hard has we need execute multiple service restarts, so this
        #       needs to work.
        if notes != "":
            notes = " (NOTE: %s)" % notes
        return_code = 0
        if self.checked_family in ("redhat", "suse"):
            if os.path.exists("/etc/rc.d/init.d/%s" % which):
                return_code = utils.subprocess_call("/sbin/service %s status > /dev/null 2>/dev/null" % which,
                                                    shell=True)
            if return_code != 0:
                status.append("service %s is not running%s" % (which, notes))
                return
        elif self.checked_family == "debian":
            # we still use /etc/init.d
            if os.path.exists("/etc/init.d/%s" % which):
                return_code = utils.subprocess_call("/etc/init.d/%s status /dev/null 2>/dev/null" % which,
                                                    shell=True)
            if return_code != 0:
                status.append("service %s is not running%s" % (which, notes))
                return
        else:
            status.append("Unknown distribution type, cannot check for running service %s" % which)
            return

    def check_iptables(self, status):
        """
        Check if iptables is running. If yes print the needed ports. This is unavailable on Debian, SUSE and CentOS7 as
        a service. However this only indicates that the way of persisting the iptable rules are persisted via other
        means.

        :param status: The status list with possible problems.
        """
        if os.path.exists("/etc/rc.d/init.d/iptables"):
            return_code = utils.subprocess_call("/sbin/service iptables status >/dev/null 2>/dev/null",
                                                shell=True)
            if return_code == 0:
                status.append("since iptables may be running, ensure 69, 80/443, and %(xmlrpc)s are unblocked"
                              % {"xmlrpc": self.settings.xmlrpc_port})

    def check_yum(self, status):
        """
        Check if the yum-stack is available. On Debian based distros this will always return without checking.

        :param status: The status list with possible problems.
        """
        # FIXME: Replace this with calls which check for the path of these tools.
        if self.checked_family == "debian":
            return

        if not os.path.exists("/usr/bin/createrepo"):
            status.append("createrepo package is not installed, needed for cobbler import and cobbler reposync, "
                          "install createrepo?")

        if not os.path.exists("/usr/bin/dnf") and not os.path.exists("/usr/bin/reposync"):
            status.append("reposync not installed, install yum-utils")

        if os.path.exists("/usr/bin/dnf") and not os.path.exists("/usr/bin/reposync"):
            status.append("reposync is not installed, install yum-utils or dnf-plugins-core")

        if not os.path.exists("/usr/bin/dnf") and not os.path.exists("/usr/bin/yumdownloader"):
            status.append("yumdownloader is not installed, install yum-utils")

        if os.path.exists("/usr/bin/dnf") and not os.path.exists("/usr/bin/yumdownloader"):
            status.append("yumdownloader is not installed, install yum-utils or dnf-plugins-core")

    def check_debmirror(self, status):
        """
        Check if debmirror is available and the config file for it exists. If the distro family is suse then this will
        pass without checking.

        :param status: The status list with possible problems.
        """
        if self.checked_family == "suse":
            return

        if not os.path.exists("/usr/bin/debmirror"):
            status.append("debmirror package is not installed, it will be required to manage debian deployments and "
                          "repositories")
        if os.path.exists("/etc/debmirror.conf"):
            with open("/etc/debmirror.conf") as f:
                re_dists = re.compile(r'@dists=')
                re_arches = re.compile(r'@arches=')
                for line in f.readlines():
                    if re_dists.search(line) and not line.strip().startswith("#"):
                        status.append("comment out 'dists' on /etc/debmirror.conf for proper debian support")
                    if re_arches.search(line) and not line.strip().startswith("#"):
                        status.append("comment out 'arches' on /etc/debmirror.conf for proper debian support")

    def check_name(self, status):
        """
        If the server name in the config file is still set to localhost automatic installations run from koan will not
        have proper kernel line parameters.

        :param status: The status list with possible problems.
        """
        if self.settings.server == "127.0.0.1":
            status.append("The 'server' field in /etc/cobbler/settings.yaml must be set to something other than localhost, "
                          "or automatic installation features will not work.  This should be a resolvable hostname or "
                          "IP for the boot server as reachable by all machines that will use it.")
        if self.settings.next_server_v4 == "127.0.0.1":
            status.append("For PXE to be functional, the 'next_server_v4' field in /etc/cobbler/settings.yaml must be set to "
                          "something other than 127.0.0.1, and should match the IP of the boot server on the PXE "
                          "network.")
        if self.settings.next_server_v6 == "::1":
            status.append("For PXE to be functional, the 'next_server_v6' field in /etc/cobbler/settings.yaml must be set to "
                          "something other than ::1, and should match the IP of the boot server on the PXE network.")

    def check_selinux(self, status):
        """
        Suggests various SELinux rules changes to run Cobbler happily with SELinux in enforcing mode.

        :param status: The status list with possible problems.
        """
        # FIXME: this method could use some refactoring in the future.
        if self.checked_family == "debian":
            return

        enabled = self.collection_mgr.api.is_selinux_enabled()
        if enabled:
            status.append("SELinux is enabled. Please review the following wiki page for details on ensuring Cobbler "
                          "works correctly in your SELinux environment:\n    "
                          "https://github.com/cobbler/cobbler/wiki/Selinux")

    def check_for_default_password(self, status):
        """
        Check if the default password of Cobbler was changed.

        :param status: The status list with possible problems.
        """
        default_pass = self.settings.default_password_crypted
        if default_pass == "$1$mF86/UHC$WvcIcX2t6crBz2onWxyac.":
            status.append("The default password used by the sample templates for newly installed machines ("
                          "default_password_crypted in /etc/cobbler/settings.yaml) is still set to 'cobbler' and should be "
                          "changed, try: \"openssl passwd -1 -salt 'random-phrase-here' 'your-password-here'\" to "
                          "generate new one")

    def check_for_unreferenced_repos(self, status):
        """
        Check if there are repositories which are not used and thus could be removed.

        :param status: The status list with possible problems.
        """
        repos = []
        referenced = []
        not_found = []
        for r in self.collection_mgr.api.repos():
            repos.append(r.name)
        for p in self.collection_mgr.api.profiles():
            my_repos = p.repos
            if my_repos != "<<inherit>>":
                referenced.extend(my_repos)
        for r in referenced:
            if r not in repos and r != "<<inherit>>":
                not_found.append(r)
        if len(not_found) > 0:
            status.append("One or more repos referenced by profile objects is no longer defined in Cobbler: %s"
                          % ", ".join(not_found))

    def check_for_unsynced_repos(self, status):
        """
        Check if there are unsynchronized repositories which need an update.

        :param status: The status list with possible problems.
        """
        need_sync = []
        for r in self.collection_mgr.repos():
            if r.mirror_locally == 1:
                lookfor = os.path.join(self.settings.webdir, "repo_mirror", r.name)
                if not os.path.exists(lookfor):
                    need_sync.append(r.name)
        if len(need_sync) > 0:
            status.append("One or more repos need to be processed by cobbler reposync for the first time before "
                          "automating installations using them: %s" % ", ".join(need_sync))

    def check_dhcpd_bin(self, status):
        """
        Check if dhcpd is installed.

        :param status: The status list with possible problems.
        """
        if not os.path.exists("/usr/sbin/dhcpd"):
            status.append("dhcpd is not installed")

    def check_dnsmasq_bin(self, status):
        """
        Check if dnsmasq is installed.

        :param status: The status list with possible problems.
        """
        return_code = utils.subprocess_call("dnsmasq --help")
        if return_code != 0:
            status.append("dnsmasq is not installed and/or in path")

    def check_bind_bin(self, status):
        """
        Check if bind is installed.

        :param status: The status list with possible problems.
        """
        return_code = utils.subprocess_call("named -v")
        # it should return something like "BIND 9.6.1-P1-RedHat-9.6.1-6.P1.fc11"
        if return_code != 0:
            status.append("named is not installed and/or in path")

    def check_for_wget_curl(self, status):
        """
        Check to make sure wget or curl is installed

        :param status: The status list with possible problems.
        """
        rc_wget = utils.subprocess_call("wget --help")
        rc_curl = utils.subprocess_call("curl --help")
        if rc_wget != 0 and rc_curl != 0:
            status.append("Neither wget nor curl are installed and/or available in $PATH. Cobbler requires that one "
                          "of these utilities be installed.")

    def check_bootloaders(self, status):
        """
        Check if network bootloaders are installed

        :param status: The status list with possible problems.
        """
        # FIXME: move zpxe.rexx to loaders

        bootloaders = {
            "menu.c32": ["/usr/share/syslinux/menu.c32",
                         "/usr/lib/syslinux/menu.c32",
                         "/var/lib/cobbler/loaders/menu.c32"],
            "pxelinux.0": ["/usr/share/syslinux/pxelinux.0",
                           "/usr/lib/syslinux/pxelinux.0",
                           "/var/lib/cobbler/loaders/pxelinux.0"],
            "efi": ["/var/lib/cobbler/loaders/grub-x86.efi",
                    "/var/lib/cobbler/loaders/grub-x86_64.efi"],
        }

        # look for bootloaders at the glob locations above
        found_bootloaders = []
        items = list(bootloaders.keys())
        for loader_name in items:
            patterns = bootloaders[loader_name]
            for pattern in patterns:
                matches = glob.glob(pattern)
                if len(matches) > 0:
                    found_bootloaders.append(loader_name)
        not_found = []

        # invert the list of what we've found so we can report on what we haven't found
        for loader_name in items:
            if loader_name not in found_bootloaders:
                not_found.append(loader_name)

        if len(not_found) > 0:
            status.append("some network boot-loaders are missing from /var/lib/cobbler/loaders. If you only want to "
                          "handle x86/x86_64 netbooting, you may ensure that you have installed a *recent* version "
                          "of the syslinux package installed and can ignore this message entirely. Files in this "
                          "directory, should you want to support all architectures, should include pxelinux.0, and"
                          "menu.c32.")

    def check_tftpd_dir(self, status):
        """
        Check if cobbler.conf's tftpboot directory exists

        :param status: The status list with possible problems.
        """
        if self.checked_family == "debian":
            return

        bootloc = self.settings.tftpboot_location
        if not os.path.exists(bootloc):
            status.append("please create directory: %(dirname)s" % {"dirname": bootloc})

    def check_ctftpd_dir(self, status):
        """
        Check if ``cobbler.conf``'s tftpboot directory exists.

        :param status: The status list with possible problems.
        """
        if self.checked_family == "debian":
            return

        bootloc = self.settings.tftpboot_location
        if not os.path.exists(bootloc):
            status.append("please create directory: %(dirname)s" % {"dirname": bootloc})

    def check_rsync_conf(self, status):
        """
        Check that rsync is enabled to autostart.

        :param status: The status list with possible problems.
        """
        if self.checked_family == "debian":
            return

        if os.path.exists("/usr/lib/systemd/system/rsyncd.service"):
            if not os.path.exists("/etc/systemd/system/multi-user.target.wants/rsyncd.service"):
                status.append("enable and start rsyncd.service with systemctl")

    def check_dhcpd_conf(self, status):
        """
        NOTE: this code only applies if Cobbler is *NOT* set to generate a ``dhcp.conf`` file.

        Check that dhcpd *appears* to be configured for pxe booting. We can't assure file correctness. Since a Cobbler
        user might have dhcp on another server, it's okay if it's not there and/or not configured correctly according
        to automated scans.

        :param status: The status list with possible problems.
        """
        if self.settings.manage_dhcp:
            return

        if os.path.exists(self.settings.dhcpd_conf):
            match_next = False
            match_file = False
            f = open(self.settings.dhcpd_conf)
            for line in f.readlines():
                if line.find("next-server") != -1:
                    match_next = True
                if line.find("filename") != -1:
                    match_file = True
            if not match_next:
                status.append("expecting next-server entry in %(file)s" % {"file": self.settings.dhcpd_conf})
            if not match_file:
                status.append("missing file: %(file)s" % {"file": self.settings.dhcpd_conf})
        else:
            status.append("missing file: %(file)s" % {"file": self.settings.dhcpd_conf})



"""
Copyright 2006-2009, Red Hat, Inc and Others
Michael DeHaan <michael.dehaan AT gmail>

This software may be freely redistributed under the terms of the GNU
general public license.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA.
"""
import copy
import enum
import fnmatch
import logging
import pprint
import re
import uuid
from typing import Any, List, Union

import yaml

from cobbler import utils, enums
from cobbler.cexceptions import CX


RE_OBJECT_NAME = re.compile(r'[a-zA-Z0-9_\-.:]*$')


class Item:
    """
    An Item is a serializable thing that can appear in a Collection
    """
    # Constants
    TYPE_NAME = "generic"
    COLLECTION_TYPE = "generic"

    @classmethod
    def __find_compare(cls, from_search: Union[str, list, dict, bool], from_obj: Union[str, list, dict, bool]):
        """
        Only one of the two parameters shall be given in this method. If you give both ``from_obj`` will be preferred.

        :param from_search: Tries to parse this str in the format as a search result string.
        :param from_obj: Tries to parse this str in the format of an obj str.
        :return: True if the comparison succeeded, False otherwise.
        :raises TypeError: In case the type of one of the two variables is wrong or could not be converted
                           intelligently.
        """
        if isinstance(from_obj, str):
            # FIXME: fnmatch is only used for string to string comparisons which should cover most major usage, if
            #        not, this deserves fixing
            from_obj_lower = from_obj.lower()
            from_search_lower = from_search.lower()
            # It's much faster to not use fnmatch if it's not needed
            if '?' not in from_search_lower and '*' not in from_search_lower and '[' not in from_search_lower:
                match = from_obj_lower == from_search_lower
            else:
                match = fnmatch.fnmatch(from_obj_lower, from_search_lower)
            return match
        else:
            if isinstance(from_search, str):
                if isinstance(from_obj, list):
                    from_search = utils.input_string_or_list(from_search)
                    for x in from_search:
                        if x not in from_obj:
                            return False
                    return True
                if isinstance(from_obj, dict):
                    (junk, from_search) = utils.input_string_or_dict(from_search, allow_multiples=True)
                    for x in list(from_search.keys()):
                        y = from_search[x]
                        if x not in from_obj:
                            return False
                        if not (y == from_obj[x]):
                            return False
                    return True
                if isinstance(from_obj, bool):
                    if from_search.lower() in ["true", "1", "y", "yes"]:
                        inp = True
                    else:
                        inp = False
                    if inp == from_obj:
                        return True
                    return False

            raise TypeError("find cannot compare type: %s" % type(from_obj))

    def __init__(self, api, is_subobject: bool = False):
        """
        Constructor.  Requires a back reference to the CollectionManager object.

        NOTE: is_subobject is used for objects that allow inheritance in their trees. This inheritance refers to
        conceptual inheritance, not Python inheritance. Objects created with is_subobject need to call their
        setter for parent immediately after creation and pass in a value of an object of the same type. Currently this
        is only supported for profiles. Subobjects blend their data with their parent objects and only require a valid
        parent name and a name for themselves, so other required options can be gathered from items further up the
        Cobbler tree.

                           distro
                               profile
                                    profile  <-- created with is_subobject=True
                                         system   <-- created as normal

        For consistency, there is some code supporting this in all object types, though it is only usable
        (and only should be used) for profiles at this time.  Objects that are children of
        objects of the same type (i.e. subprofiles) need to pass this in as True.  Otherwise, just
        use False for is_subobject and the parent object will (therefore) have a different type.

        :param api: The Cobbler API object which is used for resolving information.
        :param is_subobject: See above extensive description.
        """
        self._parent = ''
        self._depth = 0
        self._children = []
        self._ctime = 0.0
        self._mtime = 0.0
        self._uid = uuid.uuid4().hex
        self._name = ""
        self._comment = ""
        self._kernel_options: Union[dict, str] = {}
        self._kernel_options_post: Union[dict, str] = {}
        self._autoinstall_meta: Union[dict, str] = {}
        self._fetchable_files: Union[dict, str] = {}
        self._boot_files: Union[dict, str] = {}
        self._template_files = {}
        self._last_cached_mtime = 0
        self._owners: Union[list, str] = api.settings().default_ownership
        self._cached_dict = ""
        self._mgmt_classes: Union[list, str] = []
        self._mgmt_parameters: Union[dict, str] = {}
        self._conceptual_parent = None
        self._is_subobject = is_subobject

        self.logger = logging.getLogger()
        self.api = api

    def __eq__(self, other):
        """
        Comparison based on the uid for our items.

        :param other: The other Item to compare.
        :return: True if uid is equal, otherwise false.
        """
        if isinstance(other, Item):
            return self._uid == other.uid
        return False

    def _resolve(self, property_name: str) -> Any:
        """
        Resolve the ``property_name`` value in the object tree. This function traverses the tree from the object to its
        topmost parent and returns the first value that is not inherited. If the the tree does not contain a value the
        settings are consulted.

        :param property_name: The property name to resolve.
        :raises AttributeError: In case one of the objects try to inherit from a parent that does not have
                                ``property_name``.
        :return: The resolved value.
        """
        settings_name = property_name
        if property_name.startswith("proxy_url_"):
            property_name = "proxy"
        attribute = "_" + property_name

        if not hasattr(self, attribute):
            raise AttributeError("%s \"%s\" does not have property \"%s\"" % (type(self), self.name, property_name))

        attribute_value = getattr(self, attribute)
        settings = self.api.settings()

        if attribute_value == enums.VALUE_INHERITED:
            if self.parent is not None and hasattr(self.parent, property_name):
                return getattr(self.parent, property_name)
            elif hasattr(settings, settings_name):
                return getattr(settings, settings_name)
            else:
                AttributeError("%s \"%s\" inherits property \"%s\", but neither its parent nor settings have it"
                               % (type(self), self.name, property_name))

        return attribute_value

    def _resolve_dict(self, property_name: str) -> dict:
        """
        Merge the ``property_name`` dictionary of the object with the ``property_name`` of all its parents. The value
        of the child takes precedence over the value of the parent.

        :param property_name: The property name to resolve.
        :return: The merged dictionary.
        :raises AttributeError: In case the the the object had no attribute with the name :py:property_name: .
        """
        attribute = "_" + property_name

        if not hasattr(self, attribute):
            raise AttributeError("%s \"%s\" does not have property \"%s\""
                                 % (type(self), self.name, property_name))

        attribute_value = getattr(self, attribute)
        settings = self.api.settings()

        merged_dict = {}

        if self.parent is not None and hasattr(self.parent, property_name):
            merged_dict.update(getattr(self.parent, property_name))
        elif hasattr(settings, property_name):
            merged_dict.update(getattr(settings, property_name))

        if attribute_value != enums.VALUE_INHERITED:
            merged_dict.update(attribute_value)

        utils.dict_annihilate(merged_dict)

        return merged_dict

    @property
    def uid(self) -> str:
        """
        The uid is the internal unique representation of a Cobbler object. It should never be used twice, even after an
        object was deleted.

        :getter: The uid for the item. Should be unique across a running Cobbler instance.
        :setter: The new uid for the object. Should only be used by the Cobbler Item Factory.
        """
        return self._uid

    @uid.setter
    def uid(self, uid: str):
        """
        Setter for the uid of the item.

        :param uid: The new uid.
        """
        self._uid = uid

    @property
    def ctime(self) -> float:
        """
        Property which represents the creation time of the object.

        :getter: The float which can be passed to Python time stdlib.
        :setter: Should only be used by the Cobbler Item Factory.
        """
        return self._ctime

    @ctime.setter
    def ctime(self, ctime: float):
        """
        Setter for the ctime property.

        :param ctime: The time the object was created.
        :raises TypeError: In case ``ctime`` was not of type float.
        """
        if not isinstance(ctime, float):
            raise TypeError("ctime needs to be of type float")
        self._ctime = ctime

    @property
    def name(self):
        """
        Property which represents the objects name.

        :getter: The name of the object.
        :setter: Updating this has broad implications. Please try to use the ``rename()`` functionality from the
                 corresponding collection.
        """
        return self._name

    @name.setter
    def name(self, name: str):
        """
        The objects name.

        :param name: object name string
        :raises TypeError: In case ``name`` was not of type str.
        :raises ValueError: In case there were disallowed characters in the name.
        """
        if not isinstance(name, str):
            raise TypeError("name must of be type str")
        if not RE_OBJECT_NAME.match(name):
            raise ValueError("Invalid characters in name: '%s'" % name)
        self._name = name

    @property
    def comment(self) -> str:
        """
        For every object you are able to set a unique comment which will be persisted on the object.

        :getter: The comment or an emtpy string.
        :setter: The new comment for the item.
        """
        return self._comment

    @comment.setter
    def comment(self, comment: str):
        """
        Setter for the comment of the item.

        :param comment: The new comment. If ``None`` the comment will be set to an emtpy string.
        """
        self._comment = comment

    @property
    def owners(self):
        """
        This is a feature which is related to the ownership module of Cobbler which gives only specific people access
        to specific records. Otherwise this is just a cosmetic feature to allow assigning records to specific users.

        .. warning:: This is never validated against a list of existing users. Thus you can lock yourself out of a
                     record.

        :getter: Return the list of users which are currently assigned to the record.
        :setter: The list of people which should be new owners. May lock you out if you are using the ownership
                 authorization module.
        """
        return self._owners

    @owners.setter
    def owners(self, owners: list):
        """
        Setter for the ``owners`` property.

        :param owners: The new list of owners. Will not be validated for existence.
        """
        self._owners = utils.input_string_or_list(owners)

    @property
    def kernel_options(self) -> dict:
        """
        Kernel options are a space delimited list, like 'a=b c=d e=f g h i=j' or a dict.

        :getter: The parsed kernel options.
        :setter: The new kernel options as a space delimited list. May raise ``ValueError`` in case of parsing problems.
        """
        return self._resolve_dict("kernel_options")

    @kernel_options.setter
    def kernel_options(self, options):
        """
        Setter for ``kernel_options``.

        :param options: The new kernel options as a space delimited list.
        :raises ValueError: In case the values set could not be parsed successfully.
        """
        (success, value) = utils.input_string_or_dict(options, allow_multiples=True)
        if not success:
            raise ValueError("invalid kernel options")
        else:
            self._kernel_options = value

    @property
    def kernel_options_post(self) -> dict:
        """
        Post kernel options are a space delimited list, like 'a=b c=d e=f g h i=j' or a dict.

        :getter: The dictionary with the parsed values.
        :setter: Accepts str in above mentioned format or directly a dict.
        """
        return self._resolve_dict("kernel_options_post")

    @kernel_options_post.setter
    def kernel_options_post(self, options):
        """
        Setter for ``kernel_options_post``.

        :param options: The new kernel options as a space delimited list.
        :raises ValueError: In case the options could not be split successfully.
        """
        (success, value) = utils.input_string_or_dict(options, allow_multiples=True)
        if not success:
            raise ValueError("invalid post kernel options")
        else:
            self._kernel_options_post = value

    @property
    def autoinstall_meta(self) -> dict:
        """
        A comma delimited list of key value pairs, like 'a=b,c=d,e=f' or a dict.
        The meta tags are used as input to the templating system to preprocess automatic installation template files.

        :getter: The metadata or an empty dict.
        :setter: Accepts anything which can be split by :meth:`~cobbler.utils.input_string_or_dict`.
        """
        return self._resolve_dict("autoinstall_meta")

    @autoinstall_meta.setter
    def autoinstall_meta(self, options: dict):
        """
        Setter for the ``autoinstall_meta`` property.

        :param options: The new options for the automatic installation meta options.
        :raises ValueError: If splitting the value does not succeed.
        """
        (success, value) = utils.input_string_or_dict(options, allow_multiples=True)
        if not success:
            raise ValueError("invalid options given for autoinstall meta")
        else:
            self._autoinstall_meta = value

    @property
    def mgmt_classes(self) -> list:
        """
        Assigns a list of configuration management classes that can be assigned to any object, such as those used by
        Puppet's external_nodes feature.

        :getter: An empty list or the list of mgmt_classes.
        :setter: Will split this according to :meth:`~cobbler.utils.input_string_or_list`.
        """
        return self._resolve("mgmt_classes")

    @mgmt_classes.setter
    def mgmt_classes(self, mgmt_classes: list):
        """
        Setter for the ``mgmt_classes`` property.

        :param mgmt_classes: The new options for the management classes of an item.
        """
        self._mgmt_classes = utils.input_string_or_list(mgmt_classes)

    @property
    def mgmt_parameters(self) -> dict:
        """
        Parameters which will be handed to your management application (Must be a valid YAML dictionary)

        :getter: The mgmt_parameters or an empty dict.
        :setter: A YAML string which can be assigned to any object, this is used by Puppet's external_nodes feature.
        """
        return self._resolve_dict("mgmt_parameters")

    @mgmt_parameters.setter
    def mgmt_parameters(self, mgmt_parameters: Union[str, dict]):
        """
        A YAML string which can be assigned to any object, this is used by Puppet's external_nodes feature.

        :param mgmt_parameters: The management parameters for an item.
        :raises TypeError: In case the parsed YAML isn't of type dict afterwards.
        """
        if not isinstance(mgmt_parameters, (str, dict)):
            raise TypeError("mgmt_parameters must be of type str or dict")
        if isinstance(mgmt_parameters, str):
            if mgmt_parameters == enums.VALUE_INHERITED:
                self._mgmt_parameters = enums.VALUE_INHERITED
            else:
                mgmt_parameters = yaml.safe_load(mgmt_parameters)
                if not isinstance(mgmt_parameters, dict):
                    raise TypeError("Input YAML in Puppet Parameter field must evaluate to a dictionary.")
        self._mgmt_parameters = mgmt_parameters

    @property
    def template_files(self) -> dict:
        """
        File mappings for built-in configuration management

        :getter: The dictionary with name-path key-value pairs.
        :setter: A dict. If not a dict must be a str which is split by :meth:`~cobbler.utils.input_string_or_dict`.
                 Raises ``TypeError`` otherwise.
        """
        return self._template_files

    @template_files.setter
    def template_files(self, template_files: dict):
        """
        A comma seperated list of source=destination templates that should be generated during a sync.

        :param template_files: The new value for the template files which are used for the item.
        :raises ValueError: In case the conversion from non dict values was not successful.
        """
        (success, value) = utils.input_string_or_dict(template_files, allow_multiples=False)
        if not success:
            raise ValueError("template_files should be of type dict")
        else:
            self._template_files = value

    @property
    def boot_files(self) -> dict:
        """
        Files copied into tftpboot beyond the kernel/initrd

        :getter: The dictionary with name-path key-value pairs.
        :setter: A dict. If not a dict must be a str which is split by :meth:`~cobbler.utils.input_string_or_dict`.
                 Raises ``TypeError`` otherwise.
        """
        return self._resolve_dict("boot_files")

    @boot_files.setter
    def boot_files(self, boot_files: dict):
        """
        A comma separated list of req_name=source_file_path that should be fetchable via tftp.

        :param boot_files: The new value for the boot files used by the item.
        """
        (success, value) = utils.input_string_or_dict(boot_files, allow_multiples=False)
        if not success:
            raise TypeError("boot_files were handed wrong values")
        else:
            self._boot_files = value

    @property
    def fetchable_files(self) -> dict:
        """
        A comma seperated list of ``virt_name=path_to_template`` that should be fetchable via tftp or a webserver

        :getter: The dictionary with name-path key-value pairs.
        :setter: A dict. If not a dict must be a str which is split by :meth:`~cobbler.utils.input_string_or_dict`.
                 Raises ``TypeError`` otherwise.
        """
        return self._resolve_dict("fetchable_files")

    @fetchable_files.setter
    def fetchable_files(self, fetchable_files: Union[str, dict]):
        """
        Setter for the fetchable files.

        :param fetchable_files: Files which will be made available to external users.
        """
        (success, value) = utils.input_string_or_dict(fetchable_files, allow_multiples=False)
        if not success:
            raise TypeError("fetchable_files were handed wrong values")
        else:
            self._fetchable_files = value

    @property
    def depth(self) -> int:
        """
        This represents the logical depth of an object in the category of the same items. Important for the order of
        loading items from the disk and other related features where the alphabetical order is incorrect for sorting.

        :getter: The logical depth of the object.
        :setter: The new int for the logical object-depth.
        """
        return self._depth

    @depth.setter
    def depth(self, depth: int):
        """
        Setter for depth.

        :param depth: The new value for depth.
        """
        if not isinstance(depth, int):
            raise TypeError("depth needs to be of type int")
        self._depth = depth

    @property
    def mtime(self) -> float:
        """
        Represents the last modification time of the object via the API. This is not updated automagically.

        :getter: The float which can be fed into a Python time object.
        :setter: The new time something was edited via the API.
        """
        return self._mtime

    @mtime.setter
    def mtime(self, mtime: float):
        """
        Setter for the modification time of the object.

        :param mtime: The new modification time.
        """
        if not isinstance(mtime, float):
            raise TypeError("mtime needs to be of type float")
        self._mtime = mtime

    @property
    def parent(self):
        """
        This property contains the name of the logical parent of an object. In case there is not parent this return
        None.

        :getter: Returns the parent object or None if it can't be resolved via the Cobbler API.
        :setter: The name of the new logical parent.
        """
        return None

    @parent.setter
    def parent(self, parent: str):
        """
        Set the parent object for this object.

        :param parent: The new parent object. This needs to be a descendant in the logical inheritance chain.
        """

    @property
    def children(self) -> List[str]:
        """
        The list of logical children of any depth.

        :getter: An empty list in case of items which don't have logical children.
        :setter: Replace the list of children completely with the new provided one.
        """
        return []

    @children.setter
    def children(self, value):
        """
        This is an empty setter to not throw on setting it accidentally.

        :param value: The list with children names to replace the current one with.
        """
        self.logger.warning("Tried to set the children property on object \"%s\" without logical children.", self.name)

    def get_children(self, sort_list: bool = False) -> List[str]:
        """
        Get the list of children names.

        :param sort_list: If the list should be sorted alphabetically or not.
        :return: A copy of the list of children names.
        """
        result = copy.deepcopy(self.children)
        if sort_list:
            result.sort()
        return result

    @property
    def descendants(self) -> list:
        """
        Get objects that depend on this object, i.e. those that would be affected by a cascading delete, etc.

        .. note:: This is a read only property.

        :getter: This is a list of all descendants. May be empty if none exist.
        """
        results = []
        kids = self.children
        for kid in kids:
            # FIXME: Get kid objects
            grandkids = kid.descendants
            results.extend(grandkids)
        return results

    @property
    def is_subobject(self) -> bool:
        """
        Weather the object is a subobject of another object or not.

        :getter: True in case the object is a subobject, False otherwise.
        :setter: Sets the value. If this is not a bool, this will raise a ``TypeError``.
        """
        return self._is_subobject

    @is_subobject.setter
    def is_subobject(self, value: bool):
        """
        Setter for the property ``is_subobject``.

        :param value: The boolean value whether this is a subobject or not.
        :raises TypeError: In case the value was not of type bool.
        """
        if not isinstance(value, bool):
            raise TypeError("Field is_subobject of object item needs to be of type bool!")
        self._is_subobject = value

    def get_conceptual_parent(self):
        """
        The parent may just be a superclass for something like a subprofile. Get the first parent of a different type.

        :return: The first item which is conceptually not from the same type.
        """
        mtype = type(self)
        parent = self.parent
        while parent is not None:
            ptype = type(parent)
            if mtype != ptype:
                self._conceptual_parent = parent
                return parent
            parent = parent.parent
        return None

    def sort_key(self, sort_fields: list = None):
        """
        Convert the item to a dict and sort the data after specific given fields.

        :param sort_fields: The fields to sort the data after.
        :return: The sorted data.
        """
        data = self.to_dict()
        return [data.get(x, "") for x in sort_fields]

    def find_match(self, kwargs, no_errors=False):
        """
        Find from a given dict if the item matches the kv-pairs.

        :param kwargs: The dict to match for in this item.
        :param no_errors: How strict this matching is.
        :return: True if matches or False if the item does not match.
        """
        # used by find() method in collection.py
        data = self.to_dict()
        for (key, value) in list(kwargs.items()):
            # Allow ~ to negate the compare
            if value is not None and value.startswith("~"):
                res = not self.find_match_single_key(data, key, value[1:], no_errors)
            else:
                res = self.find_match_single_key(data, key, value, no_errors)
            if not res:
                return False

        return True

    def find_match_single_key(self, data, key, value, no_errors: bool = False) -> bool:
        """
        Look if the data matches or not. This is an alternative for ``find_match()``.

        :param data: The data to search through.
        :param key: The key to look for int the item.
        :param value: The value for the key.
        :param no_errors: How strict this matching is.
        :return: Whether the data matches or not.
        """
        # special case for systems
        key_found_already = False
        if "interfaces" in data:
            if key in ["mac_address", "ip_address", "netmask", "virt_bridge", "dhcp_tag", "dns_name", "static_routes",
                       "interface_type", "interface_master", "bonding_opts", "bridge_opts", "interface"]:
                key_found_already = True
                for (name, interface) in list(data["interfaces"].items()):
                    if value == name:
                        return True
                    if value is not None and key in interface:
                        if self.__find_compare(interface[key], value):
                            return True

        if key not in data:
            if not key_found_already:
                if not no_errors:
                    # FIXME: removed for 2.0 code, shouldn't cause any problems to not have an exception here?
                    # raise CX("searching for field that does not exist: %s" % key)
                    return False
            else:
                if value is not None:  # FIXME: new?
                    return False

        if value is None:
            return True
        else:
            return self.__find_compare(value, data[key])

    def dump_vars(self, formatted_output: bool = True):
        """
        Dump all variables.

        :param formatted_output: Whether to format the output or not.
        :return: The raw or formatted data.
        """
        raw = utils.blender(self.api, False, self)
        if formatted_output:
            return pprint.pformat(raw)
        else:
            return raw

    def check_if_valid(self):
        """
        Raise exceptions if the object state is inconsistent.

        :raises CX: In case the name of the item is not set.
        """
        if not self.name:
            raise CX("Name is required")

    def make_clone(self):
        """
        Must be defined in any subclass
        """
        raise NotImplementedError("Must be implemented in a specific Item")

    @classmethod
    def _remove_depreacted_dict_keys(cls, dictionary: dict):
        """
        This method does remove keys which should not be deserialized and are only there for API compatibility in
        :meth:`~cobbler.items.item.Item.to_dict`.

        :param dictionary: The dict to update
        """
        if "ks_meta" in dictionary:
            dictionary.pop("ks_meta")
        if "kickstart" in dictionary:
            dictionary.pop("kickstart")

    def from_dict(self, dictionary: dict):
        """
        Modify this object to take on values in ``dictionary``.

        :param dictionary: This should contain all values which should be updated.
        :raises AttributeError: In case during the process of setting a value for an attribute an error occurred.
        :raises KeyError: In case there were keys which could not be set in the item dictionary.
        """
        result = copy.deepcopy(dictionary)
        for key in dictionary:
            lowered_key = key.lower()
            # The following also works for child classes because self is a child class at this point and not only an
            # Item.
            if hasattr(self, "_" + lowered_key):
                try:
                    setattr(self, lowered_key, dictionary[key])
                except AttributeError as error:
                    raise AttributeError("Attribute \"%s\" could not be set!" % lowered_key) from error
                result.pop(key)
        if len(result) > 0:
            raise KeyError("The following keys supplied could not be set: %s" % result.keys())

    def to_dict(self) -> dict:
        """
        This converts everything in this object to a dictionary.

        :return: A dictionary with all values present in this object.
        """
        value = {}
        for key in self.__dict__:
            if key.startswith("_") and not key.startswith("__"):
                if key in ("_conceptual_parent", "_last_cached_mtime", "_cached_dict", "_supported_boot_loaders"):
                    continue
                new_key = key[1:].lower()
                if isinstance(self.__dict__[key], enum.Enum):
                    value[new_key] = self.__dict__[key].value
                elif new_key == "interfaces":
                    # This is the special interfaces dict. Lets fix it before it gets to the normal process.
                    serialized_interfaces = {}
                    interfaces = self.__dict__[key]
                    for interface_key in interfaces:
                        serialized_interfaces[interface_key] = interfaces[interface_key].to_dict()
                    value[new_key] = serialized_interfaces
                elif isinstance(self.__dict__[key], (list, dict)):
                    value[new_key] = copy.deepcopy(self.__dict__[key])
                else:
                    value[new_key] = self.__dict__[key]
        if "autoinstall" in value:
            value.update({"kickstart": value["autoinstall"]})
        if "autoinstall_meta" in value:
            value.update({"ks_meta": value["autoinstall_meta"]})
        return value

    def serialize(self) -> dict:
        """
        This method is a proxy for :meth:`~cobbler.items.item.Item.to_dict` and contains additional logic for
        serialization to a persistent location.

        :return: The dictionary with the information for serialization.
        """
        keys_to_drop = ["kickstart", "ks_meta", "remote_grub_kernel", "remote_grub_initrd"]
        result = self.to_dict()
        for key in keys_to_drop:
            result.pop(key, "")
        return result

    def deserialize(self, item_dict: dict):
        """
        This is currently a proxy for :py:meth:`~cobbler.items.item.Item.from_dict` .

        :param item_dict: The dictionary with the data to deserialize.
        """
        self.from_dict(item_dict)



"""
Copyright 2021 Yuriy Chelpanov
Yuriy Chelpanov <yuriy.chelpanov@gmail.com>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301  USA
"""

from cobbler.items import menu
from cobbler.cobbler_collections import collection
from cobbler import utils
from cobbler.cexceptions import CX


class Menus(collection.Collection):
    """
    A menu represents an element of the hierarchical boot menu.
    """

    @staticmethod
    def collection_type() -> str:
        return "menu"

    @staticmethod
    def collection_types() -> str:
        return "menus"

    def factory_produce(self, api, item_dict):
        """
        Return a Menu forged from item_dict

        :param api: The cobblerd API.
        :param item_dict: The seed data.
        :return: A new menu instance.
        """
        new_menu = menu.Menu(api)
        new_menu.from_dict(item_dict)
        return new_menu

    def remove(self, name: str, with_delete: bool = True, with_sync: bool = True, with_triggers: bool = True,
               recursive: bool = False):
        """
        Remove element named 'name' from the collection

        :param name: The name of the menu
        :param with_delete: TODO
        :param with_sync: TODO
        :param with_triggers: TODO
        :param recursive: TODO
        :raises CX: Raised in case you want to delete a none existing menu.
        """
        name = name.lower()
        for profile in self.api.profiles():
            if profile.menu and profile.menu.lower() == name:
                profile.menu = ""
        for image in self.api.images():
            if image.menu and image.menu.lower() == name:
                image.menu = ""

        obj = self.find(name=name)
        if obj is not None:
            if recursive:
                kids = obj.get_children()
                for kid in kids:
                    self.remove(kid, with_delete=with_delete, with_sync=False, recursive=recursive)

            if with_delete:
                if with_triggers:
                    utils.run_triggers(self.api, obj, "/var/lib/cobbler/triggers/delete/menu/pre/*", [])
            self.lock.acquire()
            try:
                del self.listing[name]
            finally:
                self.lock.release()
            self.collection_mgr.serialize_delete(self, obj)
            if with_delete:
                if with_triggers:
                    utils.run_triggers(self.api, obj, "/var/lib/cobbler/triggers/delete/menu/post/*", [])
                    utils.run_triggers(self.api, obj, "/var/lib/cobbler/triggers/change/*", [])
                if with_sync:
                    lite_sync = self.api.get_sync()
                    lite_sync.remove_single_menu()
            return
        raise CX("cannot delete an object that does not exist: %s" % name)



import os
from xmlrpc.client import Fault

import pytest


@pytest.mark.usefixtures("cobbler_xmlrpc_base")
class TestDistro:
    """
    Test remote calls related to distros.
    """

    def test_get_distros(self, remote, token):
        """
        Test: get distros
        """
        # Arrange --> Nothing to arrange

        # Act
        result = remote.get_distros(token)

        # Assert
        assert result == []

    @pytest.mark.parametrize("field_name,field_value", [
        ("arch", "i386"),
        ("breed", "debian"),
        ("breed", "freebsd"),
        ("breed", "redhat"),
        ("breed", "suse"),
        ("breed", "ubuntu"),
        ("breed", "unix"),
        ("breed", "vmware"),
        ("breed", "windows"),
        ("breed", "xen"),
        ("breed", "generic"),
        ("comment", "test comment"),
        ("initrd", ""),
        ("name", "testdistro0"),
        ("kernel", ""),
        ("kernel_options", "a=1 b=2 c=3 c=4 c=5 d e"),
        ("kernel_options_post", "a=1 b=2 c=3 c=4 c=5 d e"),
        ("autoinstall_meta", "a=1 b=2 c=3 c=4 c=5 d e"),
        ("mgmt_classes", "one two three"),
        ("os_version", "rhel4"),
        ("owners", "user1 user2 user3"),
        ("boot_loaders", "pxe ipxe grub")
    ])
    def test_create_distro_positive(self, remote, token, create_kernel_initrd, fk_kernel, fk_initrd, field_name,
                                    field_value):
        """
        Test: create/edit a distro with valid values
        """
        # Arrange --> Nothing to do.
        folder = create_kernel_initrd(fk_kernel, fk_initrd)
        distro = remote.new_distro(token)
        remote.modify_distro(distro, "name", "testdistro", token)

        # Act
        if field_name == "kernel":
            field_value = os.path.join(folder, fk_kernel)
        if field_name == "initrd":
            field_value = os.path.join(folder, fk_initrd)
        result = remote.modify_distro(distro, field_name, field_value, token)

        # Assert
        assert result

    @pytest.mark.parametrize("field_name,field_value", [
        ("arch", "badarch"),
        ("breed", "badbreed"),
        # ("boot_loader", "badloader") FIXME: This does not raise but did in the past
    ])
    def test_create_distro_negative(self, remote, token, field_name, field_value):
        """
        Test: create/edit a distro with invalid values
        """
        # Arrange
        distro = remote.new_distro(token)
        remote.modify_distro(distro, "name", "testdistro0", token)

        # Act & Assert
        try:
            remote.modify_distro(distro, field_name, field_value, token)
        except Fault:
            assert True
        else:
            pytest.fail("Bad field did not raise an exception!")

    @pytest.mark.usefixtures("create_testdistro", "remove_testdistro")
    def test_get_distro(self, remote, fk_initrd, fk_kernel):
        """
        Test: get a distro object
        """
        # Arrange --> Done in fixture

        # Act
        distro = remote.get_distro("testdistro0")

        # Assert
        assert distro.get("name") == "testdistro0"
        assert fk_initrd in distro.get("initrd")
        assert fk_kernel in distro.get("kernel")

    def test_get_system(self, remote):
        """
        Test: get a system object
        """
        # Arrange --> There should be no system present. --> Nothing to Init.

        # Act
        system = remote.get_system("testsystem0")

        # Assert
        assert system is "~"

    def test_find_distro(self, remote, token):
        """
        Test: find a distro object
        """
        # Arrange --> No distros means no setup

        # Act
        result = remote.find_distro({"name": "testdistro0"}, token)

        # Assert
        assert result == []

    @pytest.mark.usefixtures("create_testdistro", "remove_testdistro")
    def test_copy_distro(self, remote, token):
        """
        Test: copy a distro object
        """
        # Arrange --> Done in the fixture

        # Act
        distro = remote.get_item_handle("distro", "testdistro0", token)
        result = remote.copy_distro(distro, "testdistrocopy", token)

        # Assert
        assert result

        # Cleanup --> Plus fixture
        remote.remove_distro("testdistrocopy", token)

    @pytest.mark.usefixtures("create_testdistro")
    def test_rename_distro(self, remote, token):
        """
        Test: rename a distro object
        """
        # Arrange
        distro = remote.get_item_handle("distro", "testdistro0", token)

        # Act
        result = remote.rename_distro(distro, "testdistro1", token)

        # Assert
        assert result

        # Cleanup
        remote.remove_distro("testdistro1", token)

    def test_remove_distro(self, remote, token):
        """
        Test: remove a distro object
        """
        # Arrange
        # TODO: Verify why the test passes without the fixture for creating the distro!

        # Act
        result = remote.remove_distro("testdistro0", token)

        # Assert
        assert result
